import numpy as np
import pytest
import xarray as xr

from climpred import set_options
from climpred.constants import (
    BIAS_CORRECTION_BIAS_CORRECTION_METHODS,
    GROUPBY_SEASONALITIES,
    INTERNAL_BIAS_CORRECTION_METHODS,
    XCLIM_BIAS_CORRECTION_METHODS,
)
from climpred.options import OPTIONS
from climpred.testing import assert_PredictionEnsemble

BIAS_CORRECTION_METHODS = (
    BIAS_CORRECTION_BIAS_CORRECTION_METHODS + INTERNAL_BIAS_CORRECTION_METHODS
)
BIAS_CORRECTION_METHODS.remove("normal_mapping")
BIAS_CORRECTION_METHODS.remove("gamma_mapping")
# fails with these conftest files somehow
XCLIM_BIAS_CORRECTION_METHODS.remove("LOCI")
XCLIM_BIAS_CORRECTION_METHODS.remove("PrincipalComponents")


def _adjust_metric_kwargs(metric_kwargs=None, how=None, he=None):
    if metric_kwargs is None:
        metric_kwargs = {}
    if how in ["LOCI"] and "thresh" not in metric_kwargs:
        v = list(he.data_vars)[0]
        metric_kwargs["thresh"] = he.get_observations().quantile(0.1)[v]
    return metric_kwargs


@pytest.mark.parametrize("how", BIAS_CORRECTION_METHODS)
def test_remove_bias_difference_seasonality(hindcast_recon_1d_mm, how):
    """Test HindcastEnsemble.remove_bias yields different results for different seasonality settings."""
    verify_kwargs = dict(
        metric="rmse", dim="init", comparison="e2o", alignment="same_inits", skipna=True
    )
    hindcast = hindcast_recon_1d_mm.isel(lead=range(3))
    v = "SST"

    bias_reduced_skill = []
    seasonalities = GROUPBY_SEASONALITIES
    for seasonality in seasonalities:
        with set_options(seasonality=seasonality):
            hindcast_rb = hindcast.remove_bias(
                how=how, alignment=verify_kwargs["alignment"], cv=False
            )

            bias_reduced_skill.append(hindcast_rb.verify(**verify_kwargs)[v])
    bias_reduced_skill = xr.concat(bias_reduced_skill, "seasonality").assign_coords(
        seasonality=seasonalities
    )

    # check not identical
    for s in seasonalities:
        assert bias_reduced_skill.sel(seasonality=s).notnull().all()
        for s2 in seasonalities:
            if s != s2:
                assert (
                    bias_reduced_skill.sel(seasonality=[s, s2])
                    .diff("seasonality")
                    .notnull()
                    .any()
                )


@pytest.mark.parametrize("cv", [False, "LOO"])
@pytest.mark.parametrize("seasonality", GROUPBY_SEASONALITIES)
@pytest.mark.parametrize("how", BIAS_CORRECTION_METHODS)
@pytest.mark.parametrize(
    "alignment", ["same_inits", "maximize"]
)  # same_verifs  # no overlap here for same_verifs
def test_remove_bias(hindcast_recon_1d_mm, alignment, how, seasonality, cv):
    """Test remove mean bias, ensure than skill doesnt degrade and keeps attrs."""

    def check_hindcast_coords_maintained_except_init(hindcast, hindcast_bias_removed):
        # init only slighty cut due to alignment
        for c in hindcast.coords:
            if c == "init":
                assert hindcast.coords[c].size >= hindcast_bias_removed.coords[c].size
            else:
                assert hindcast.coords[c].size == hindcast_bias_removed.coords[c].size

    with set_options(seasonality=seasonality):
        metric = "rmse"
        dim = "init"
        comparison = "e2o"
        hindcast = hindcast_recon_1d_mm.isel(lead=range(3))
        hindcast._datasets["initialized"].attrs["test"] = "test"
        hindcast._datasets["initialized"]["SST"].attrs["units"] = "test_unit"
        verify_kwargs = dict(
            metric=metric,
            alignment=alignment,
            dim=dim,
            comparison=comparison,
            keep_attrs=True,
        )

        # add how bias
        if "additive" in how:
            with xr.set_options(keep_attrs=True):
                hindcast._datasets["observations"] = (
                    hindcast._datasets["observations"] + 0.1
                )
        elif "multiplicative" in how:
            with xr.set_options(keep_attrs=True):
                hindcast._datasets["observations"] = (
                    hindcast._datasets["observations"] * 1.1
                )

        biased_skill = hindcast.verify(**verify_kwargs)

        hindcast_bias_removed = hindcast.remove_bias(
            how=how, alignment=alignment, cv=False
        )

        check_hindcast_coords_maintained_except_init(hindcast, hindcast_bias_removed)

        bias_removed_skill = hindcast_bias_removed.verify(**verify_kwargs)

        seasonality = OPTIONS["seasonality"]
        if cv:
            hindcast_bias_removed_properly = hindcast.remove_bias(
                how=how, cv="LOO", alignment=alignment
            )
            check_hindcast_coords_maintained_except_init(
                hindcast, hindcast_bias_removed_properly
            )

            bias_removed_skill_properly = hindcast_bias_removed_properly.verify(
                **verify_kwargs
            )
            # checks
            assert seasonality not in bias_removed_skill_properly.coords
            assert (biased_skill > bias_removed_skill_properly).all()
            assert (bias_removed_skill_properly >= bias_removed_skill).all()

        assert (biased_skill > bias_removed_skill).all()
        assert seasonality not in bias_removed_skill.coords
        # keeps data_vars attrs
        for v in hindcast_bias_removed.get_initialized().data_vars:
            if cv:
                assert (
                    hindcast_bias_removed_properly.get_initialized()[v].attrs
                    == hindcast.get_initialized()[v].attrs
                )
            assert (
                hindcast_bias_removed.get_initialized()[v].attrs
                == hindcast.get_initialized()[v].attrs
            )
        # keeps dataset attrs
        if cv:
            assert (
                hindcast_bias_removed_properly.get_initialized().attrs
                == hindcast.get_initialized().attrs
            )
        assert (
            hindcast_bias_removed.get_initialized().attrs
            == hindcast.get_initialized().attrs
        )
        # keep lead attrs
        assert (
            hindcast_bias_removed.get_initialized().lead.attrs
            == hindcast.get_initialized().lead.attrs
        )


@pytest.mark.parametrize(
    "alignment", ["same_inits", "maximize"]
)  # same_verifs  # no overlap here for same_verifs
@pytest.mark.parametrize("seasonality", GROUPBY_SEASONALITIES)
@pytest.mark.parametrize("how", BIAS_CORRECTION_METHODS)
def test_monthly_leads_remove_bias_LOO(
    hindcast_NMME_Nino34, how, seasonality, alignment
):
    """Get different HindcastEnsemble depending on CV or not."""
    with set_options(seasonality=seasonality):
        he = (
            hindcast_NMME_Nino34.isel(lead=[0, 1])
            .isel(model=2, drop=True)
            .sel(init=slice("2005", "2006"))
        )
        assert not he.remove_bias(
            how=how, alignment=alignment, cv=False, train_test_split="unfair"
        ).equals(
            he.remove_bias(
                how=how, alignment=alignment, cv="LOO", train_test_split="unfair-cv"
            )
        )


@pytest.mark.slow
@pytest.mark.parametrize("alignment", ["same_inits", "maximize", "same_verifs"])
@pytest.mark.parametrize("seasonality", ["month", "season"])
@pytest.mark.parametrize("how", BIAS_CORRECTION_METHODS)
def test_remove_bias_unfair_artificial_skill_over_fair(
    hindcast_NMME_Nino34, how, seasonality, alignment
):
    """Show how method unfair better skill than fair."""
    verify_kwargs = dict(
        metric="rmse", comparison="e2o", dim="init", alignment=alignment, skipna=True
    )

    with set_options(seasonality=seasonality):
        he = (
            hindcast_NMME_Nino34.sel(lead=[4, 5])
            .sel(model="GEM-NEMO")
            .dropna("member", how="all")
            .sel(init=slice("2000", "2009"))
        )
        v = "sst"

        print("\n unfair \n")
        he_unfair = he.remove_bias(
            how=how,
            alignment=alignment,
            train_test_split="unfair",
        )
        unfair_skill = he_unfair.verify(**verify_kwargs)

        print("\n unfair-cv \n")
        he_unfair_cv = he.remove_bias(
            how=how,
            alignment=alignment,
            train_test_split="unfair-cv",
            cv="LOO",
        )
        unfair_cv_skill = he_unfair_cv.verify(**verify_kwargs)

        print("\n fair \n")
        kw = (
            dict(train_time=slice("2000", "2004"))
            if alignment == "same_verifs"
            else dict(train_init=slice("2000", "2004"))
        )
        he_fair = he.remove_bias(
            how=how,
            alignment=alignment,
            train_test_split="fair",
            **kw,
        )

        fair_skill = he_fair.verify(**verify_kwargs)

        assert not unfair_skill[v].isnull().all()
        assert not fair_skill[v].isnull().all()

        # allow 1 or 20% margin (worse skill)
        f = 1.2 if how in ["modified_quantile", "basic_quantile"] else 1.01
        assert (fair_skill * f > unfair_skill)[v].all(), print(
            fair_skill[v], unfair_skill[v]
        )
        print("checking unfair-cv")
        assert not unfair_cv_skill[v].isnull().all()
        assert (fair_skill * f > unfair_cv_skill)[v].all(), print(
            fair_skill[v], unfair_cv_skill[v]
        )


@pytest.mark.slow
@pytest.mark.parametrize("alignment", ["same_inits", "maximize", "same_verifs"])
@pytest.mark.parametrize("seasonality", ["month", None])
@pytest.mark.parametrize("how", XCLIM_BIAS_CORRECTION_METHODS)
def test_remove_bias_unfair_artificial_skill_over_fair_xclim(
    hindcast_NMME_Nino34, how, seasonality, alignment
):
    """Show how method unfair better skill than fair in xclim methods."""
    try:
        he = (
            hindcast_NMME_Nino34.sel(lead=[4, 5])
            .sel(model="GEM-NEMO")
            .dropna("member", how="all")
            .sel(init=slice("2000", "2009"))
        )
        v = "sst"

        verify_kwargs = dict(
            metric="rmse",
            comparison="e2o",
            dim="init",
            alignment=alignment,
            skipna=True,
        )

        group = "time"
        if seasonality is not None:
            group = f"{group}.{seasonality}"

        print("\n unfair \n")

        metric_kwargs = _adjust_metric_kwargs(metric_kwargs=None, how=how, he=he)

        he_unfair = he.remove_bias(
            how=how,
            alignment=alignment,
            group=group,
            train_test_split="unfair",
            **metric_kwargs,
        )

        unfair_skill = he_unfair.verify(**verify_kwargs)

        print("\n unfair-cv \n")
        if how not in ["DetrendedQuantileMapping", "EmpiricalQuantileMapping"]:
            he_unfair_cv = he.remove_bias(
                how=how,
                alignment=alignment,
                group=group,
                train_test_split="unfair-cv",
                cv="LOO",
                **metric_kwargs,
            )
            unfair_cv_skill = he_unfair_cv.verify(**verify_kwargs)

        print("\n fair \n")
        kw = (
            dict(train_time=slice("2000", "2004"))
            if alignment == "same_verifs"
            else dict(train_init=slice("2000", "2004"))
        )
        he_fair = he.remove_bias(
            how=how,
            alignment=alignment,
            group=group,
            train_test_split="fair",
            **kw,
            **metric_kwargs,
        )

        fair_skill = he_fair.verify(**verify_kwargs)

        assert not unfair_skill[v].isnull().all()
        assert not fair_skill[v].isnull().all()
        if how not in ["LOCI"]:
            assert (fair_skill > unfair_skill)[v].all(), print(
                fair_skill[v], unfair_skill[v]
            )
            print("checking unfair-cv")
            if how not in ["DetrendedQuantileMapping", "EmpiricalQuantileMapping"]:
                assert not unfair_cv_skill[v].isnull().all()
                assert (fair_skill > unfair_cv_skill)[v].all(), print(
                    fair_skill[v], unfair_cv_skill[v]
                )

    except np.linalg.LinAlgError:  # PrincipalComponents
        print(f"np.linalg.LinAlgError: {how}")
        pass


def test_remove_bias_xclim_grouper_diff(
    hindcast_NMME_Nino34,
):
    """Test remove_bias(how='xclim_method') is sensitive to grouper"""
    alignment = "same_init"
    how = "DetrendedQuantileMapping"
    he = (
        hindcast_NMME_Nino34.sel(lead=[4, 5])
        .sel(model="GEM-NEMO")
        .dropna("member", how="all")
        .sel(init=slice("2000", "2001"))
    )

    he_time = he.remove_bias(
        how=how,
        alignment=alignment,
        group="time",
        train_test_split="unfair",
    )

    he_init = he.remove_bias(
        how=how,
        alignment=alignment,
        group="init",
        train_test_split="unfair",
    )

    he_time_month = he.remove_bias(
        how=how,
        alignment=alignment,
        group="time.month",
        train_test_split="unfair",
    )

    assert not he_time_month.equals(he_time)
    xr.testing.assert_equal(he_init.get_initialized(), he_time.get_initialized())


def test_remove_bias_xclim_adjust_kwargs_diff(
    hindcast_NMME_Nino34,
):
    """Test remove_bias(how='xclim_method') is sensitive to adjust_kwargs"""
    alignment = "same_init"
    how = "EmpiricalQuantileMapping"
    he = (
        hindcast_NMME_Nino34.sel(lead=[4, 5])
        .sel(model="GEM-NEMO")
        .dropna("member", how="all")
        .sel(init=slice("2000", "2001"))
    )

    he_interp_linear = he.remove_bias(
        how=how,
        alignment=alignment,
        group="time",
        train_test_split="unfair",
        interp="linear",
    )

    he_interp_nearest = he.remove_bias(
        how=how,
        alignment=alignment,
        group="init",
        train_test_split="unfair",
        interp="nearest",
    )

    assert not he_interp_nearest.equals(he_interp_linear)


def test_remove_bias_xclim_kwargs(hindcast_NMME_Nino34):
    """Testing kwargs are used."""
    he = (
        hindcast_NMME_Nino34.sel(lead=[4, 5])
        .sel(model="GEM-NEMO")
        .dropna("member", how="all")
        .sel(init=slice("2000", "2001"))
    )
    hind_kw = he.remove_bias(
        how="DetrendedQuantileMapping",
        alignment="same_inits",
        train_test_split="unfair",
        group="time.month",
        nquantiles=10,
    )
    hind = he.remove_bias(
        how="DetrendedQuantileMapping",
        alignment="same_inits",
        train_test_split="unfair",
        group="time.month",
    )
    assert not hind_kw.equals(hind)


def test_remove_bias_group(hindcast_NMME_Nino34):
    """Test group=None equals not providing group."""
    he = (
        hindcast_NMME_Nino34.sel(lead=[4, 5])
        .sel(model="GEM-NEMO")
        .dropna("member", how="all")
        .sel(init=slice("2000", "2001"))
    )
    hind_None = he.remove_bias(
        how="DetrendedQuantileMapping",
        alignment="same_inits",
        train_test_split="unfair",
        group=None,
    )
    hind_no = he.remove_bias(
        how="DetrendedQuantileMapping",
        alignment="same_inits",
        train_test_split="unfair",
    )
    assert hind_no.equals(hind_None)


def test_remove_bias_compare_scaling_and_mean(hindcast_recon_1d_dm):
    """Compare Scaling and additive_mean to be similar"""
    he = hindcast_recon_1d_dm.isel(lead=[0, 1]).isel(init=slice(None, 366 * 2))
    hind_scaling = he.remove_bias(
        how="Scaling",
        kind="+",
        alignment="same_inits",
        train_test_split="unfair",
        group="time.dayofyear",
    )
    with set_options(seasonality="dayofyear"):
        hind_mean = hindcast_recon_1d_dm.remove_bias(
            how="additive_mean",
            alignment="same_inits",
            train_test_split="unfair",
        )
    assert (
        (hind_scaling - hind_mean).get_initialized().mean(["member", "init"]) < 0.02
    ).SST.all()


def test_remove_bias_errors(hindcast_NMME_Nino34):
    """Test remove_bias error messaging."""
    how = "additive_mean"
    he = (
        hindcast_NMME_Nino34.sel(lead=[4, 5])
        .sel(model="GEM-NEMO")
        .dropna("member", how="all")
        .sel(init=slice("2000", "2009"))
    )

    with pytest.raises(ValueError, match="please provide `train_init`"):
        he.remove_bias(how=how, alignment="same_inits", train_test_split="fair")

    with pytest.raises(ValueError, match="please provide `train_init`"):
        he.remove_bias(
            how=how, alignment="same_inits", train_test_split="fair", train_init=2000
        )

    with pytest.raises(ValueError, match="please provide `train_time`"):
        he.remove_bias(how=how, alignment="same_verif", train_test_split="fair")

    with pytest.raises(ValueError, match="please provide `train_time`"):
        he.remove_bias(
            how=how, alignment="same_verif", train_test_split="fair", train_time=2000
        )

    with pytest.raises(ValueError, match="Please provide `cv="):
        he.remove_bias(how=how, alignment="same_verif", train_test_split="unfair-cv")

    with pytest.raises(NotImplementedError, match="please choose from"):
        he.remove_bias(how="new", alignment="same_verif", train_test_split="unfair-cv")

    for tts in ["fair-sliding", "fair-all"]:
        with pytest.raises(
            NotImplementedError, match="Please choose `train_test_split` from"
        ):
            he.remove_bias(how="new", alignment="same_verif", train_test_split="tts")



import logging
from datetime import datetime


def log_compute_hindcast_header(metric, comparison, dim, alignment, reference):
    """Add header to the log for a `compute_hindcast` instance."""
    logging.info(
        f"`compute_hindcast` for metric {metric.name}, "
        f"comparison {comparison.name}, dim {dim}, alignment {alignment} and reference {reference} at "
        f"{str(datetime.now())}\n"
        f"++++++++++++++++++++++++++++++++++++++++++++++++"
    )


def log_compute_hindcast_inits_and_verifs(
    dim, lead, inits, verif_dates, reference=None
):
    """At each lead, log the inits and verification dates being used in computations."""
    if reference is None:
        reference = "initialized"
    logging.info(
        f"{reference} | lead: {str(lead).zfill(2)} | "
        # This is the init-sliced forecast, thus displaying actual
        # initializations.
        f"inits: {inits[lead].min().values}"
        f"-{inits[lead].max().values} | "
        # This is the verification window, thus displaying the
        # verification dates.
        f"verifs: {verif_dates[lead].min()}"
        f"-{verif_dates[lead].max()}"
    )
    init_output = [
        f"{i.year}-{str(i.month).zfill(2)}-{str(i.day).zfill(2)}"
        for i in inits[lead].values
    ]
    verif_output = [
        f"{i.year}-{str(i.month).zfill(2)}-{str(i.day).zfill(2)}"
        for i in verif_dates[lead].values
    ]
    logging.debug(f"\ninits: {init_output}" f"\nverifs: {verif_output}")



import hashlib
import os as _os
import urllib
from urllib.request import urlretrieve as _urlretrieve

from xarray.backends.api import open_dataset as _open_dataset

_default_cache_dir = _os.sep.join(("~", ".climpred_data"))

aliases = [
    "MPI-control-1D",
    "MPI-control-3D",
    "MPI-PM-DP-1D",
    "MPI-PM-DP-3D",
    "CESM-DP-SST",
    "CESM-DP-SSS",
    "CESM-DP-SST-3D",
    "CESM-LE",
    "MPIESM_miklip_baseline1-hind-SST-global",
    "MPIESM_miklip_baseline1-hist-SST-global",
    "MPIESM_miklip_baseline1-assim-SST-global",
    "ERSST",
    "FOSI-SST",
    "FOSI-SSS",
    "FOSI-SST-3D",
    "GMAO-GEOS-RMM1",
    "RMM-INTERANN-OBS",
    "ECMWF_S2S_Germany",
    "Observations_Germany",
    "NMME_hindcast_Nino34_sst",
    "NMME_OIv2_Nino34_sst",
]
true_file_names = [
    "PM_MPI-ESM-LR_control",
    "PM_MPI-ESM-LR_control3d",
    "PM_MPI-ESM-LR_ds",
    "PM_MPI-ESM-LR_ds3d",
    "CESM-DP-LE.SST.global",
    "CESM-DP-LE.SSS.global",
    "CESM-DP-LE.SST.eastern_pacific",
    "CESM-LE.global_mean.SST.1955-2015",
    "MPIESM_miklip_baseline1-hind-SST-global",
    "MPIESM_miklip_baseline1-hist-SST-global",
    "MPIESM_miklip_baseline1-assim-SST-global",
    "ERSSTv4.global.mean",
    "FOSI.SST.global",
    "FOSI.SSS.global",
    "FOSI.SST.eastern_pacific",
    "GMAO-GEOS-V2p1.RMM1",
    "RMM1.observed.interannual.1974-06.2017-07",
    "ECMWF_S2S_Germany",
    "Observations_Germany",
    "NMME_hindcast_Nino34_sst",
    "NMME_Reyn_SmithOIv2_Nino34_sst",
]
file_descriptions = [
    "area averages for the MPI control run of SST/SSS.",
    "lat/lon/time for the MPI control run of SST/SSS.",
    "perfect model decadal prediction ensemble area averages of SST/SSS/AMO.",
    "perfect model decadal prediction ensemble lat/lon/time of SST/SSS/AMO.",
    "hindcast decadal prediction ensemble of global mean SSTs.",
    "hindcast decadal prediction ensemble of global mean SSS.",
    "hindcast decadal prediction ensemble of eastern Pacific SSTs.",
    "uninitialized ensemble of global mean SSTs.",
    "hindcast initialized ensemble of global mean SSTs",
    "uninitialized ensemble of global mean SSTs",
    "assimilation in MPI-ESM of global mean SSTs",
    "observations of global mean SSTs.",
    "reconstruction of global mean SSTs.",
    "reconstruction of global mean SSS.",
    "reconstruction of eastern Pacific SSTs",
    "daily RMM1 from the GMAO-GEOS-V2p1 model for SubX",
    "observed RMM with interannual variablity included",
    "S2S ECMWF on-the-fly hindcasts from the S2S Project for Germany",
    "CPC/ERA5 observations for S2S forecasts over Germany",
    "monthly multi-member hindcasts of sea-surface temperature averaged over the Nino3.4 region from the NMME project from IRIDL",
    "monthly Reyn_SmithOIv2 sea-surface temperature observations averaged over the Nino3.4 region",
]

FILE_ALIAS_DICT = dict(zip(aliases, true_file_names))
FILE_DESCRIPTIONS = dict(zip(aliases, file_descriptions))


def _file_md5_checksum(fname):
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        hash_md5.update(f.read())
    return hash_md5.hexdigest()


def _get_datasets():
    """Prints out available datasets for the user to load if no args are
    given."""
    for key in FILE_DESCRIPTIONS.keys():
        print(f"'{key}': {FILE_DESCRIPTIONS[key]}")


def _cache_all():
    """Cache all datasets for pytest -n 4 woth pytest-xdist."""
    for d in aliases:
        load_dataset(d)


def _initialize_proxy(proxy_dict):
    """Opens a proxy for firewalled servers so that the downloads can go
    through.

    Args:
        proxy_dict (dictionary): Keys are either 'http' or 'https' and
            values are the proxy server.

    Ref: https://stackoverflow.com/questions/22967084/
         urllib-request-urlretrieve-with-proxy
    """
    proxy = urllib.request.ProxyHandler(proxy_dict)
    opener = urllib.request.build_opener(proxy)
    urllib.request.install_opener(opener)


def load_dataset(
    name=None,
    cache=True,
    cache_dir=_default_cache_dir,
    github_url="https://github.com/pangeo-data/climpred-data",
    branch="master",
    extension=None,
    proxy_dict=None,
    **kws,
):
    """Load example data or a mask from an online repository.

    Args:
        name: (str, default None) Name of the netcdf file containing the
              dataset, without the .nc extension. If None, this function
              prints out the available datasets to import.
        cache_dir: (str, optional) The directory in which to search
                   for and cache the data.
        cache: (bool, optional) If True, cache data locally for use on later
               calls.
        github_url: (str, optional) Github repository where the data is stored.
        branch: (str, optional) The git branch to download from.
        extension: (str, optional) Subfolder within the repository where the
                   data is stored.
        proxy_dict: (dict, optional) Dictionary with keys as either 'http' or
                    'https' and values as the proxy server. This is useful
                    if you are on a work computer behind a firewall and need
                    to use a proxy out to download data.
        kws: (dict, optional) Keywords passed to xarray.open_dataset

    Returns:
        The desired xarray dataset.

    Examples:
        >>> from climpred.tutorial import load_dataset
        >>> proxy_dict = {'http': '127.0.0.1'}
        >>> ds = load_dataset('FOSI-SST', cache=False, proxy_dict=proxy_dict)
    """
    if name is None:
        return _get_datasets()

    if proxy_dict is not None:
        _initialize_proxy(proxy_dict)

    # https://stackoverflow.com/questions/541390/extracting-extension-from-
    # filename-in-python
    # Allows for generalized file extensions.
    name, ext = _os.path.splitext(name)
    if not ext.endswith(".nc"):
        ext += ".nc"

    # use aliases
    if name in FILE_ALIAS_DICT.keys():
        name = FILE_ALIAS_DICT[name]
    longdir = _os.path.expanduser(cache_dir)
    fullname = name + ext
    localfile = _os.sep.join((longdir, fullname))
    md5name = name + ".md5"
    md5file = _os.sep.join((longdir, md5name))

    if not _os.path.exists(localfile):
        # This will always leave this directory on disk.
        # May want to add an option to remove it.
        if not _os.path.isdir(longdir):
            _os.mkdir(longdir)

        if extension is not None:
            url = "/".join((github_url, "raw", branch, extension, fullname))
            _urlretrieve(url, localfile)
            url = "/".join((github_url, "raw", branch, extension, md5name))
            _urlretrieve(url, md5file)
        else:
            url = "/".join((github_url, "raw", branch, fullname))
            _urlretrieve(url, localfile)
            url = "/".join((github_url, "raw", branch, md5name))
            _urlretrieve(url, md5file)

        localmd5 = _file_md5_checksum(localfile)
        with open(md5file, "r") as f:
            remotemd5 = f.read()
        if localmd5 != remotemd5:
            _os.remove(localfile)
            msg = """
            Try downloading the file again. There was a confliction between
            your local .md5 file compared to the one in the remote repository,
            so the local copy has been removed to resolve the issue.
            """
            raise IOError(msg)

    ds = _open_dataset(localfile, **kws)

    if not cache:
        ds = ds.load()
        _os.remove(localfile)
    return ds



# for general checks of climpred-required dimensions
CLIMPRED_ENSEMBLE_DIMS = ["init", "member", "lead"]
# corresponding CF-complying standard_names from http://cfconventions.org/Data/cf-standard-names/current/build/cf-standard-name-table.html to rename from
CF_STANDARD_NAMES = {
    "init": "forecast_reference_time",
    "member": "realization",
    "lead": "forecast_period",
}
CF_LONG_NAMES = {
    "init": "Initialization",
    "member": "Member",
    "lead": "Lead",
}
CLIMPRED_DIMS = CLIMPRED_ENSEMBLE_DIMS + ["time"]

# List of frequencies to check to infer time series stride
FREQ_LIST_TO_INFER_STRIDE = ["day", "month", "year"]

# calendar type for perfect-model (PM) (needed for bootstrapping_uninit)
# Leap also works, but changing Leap,NoLeap fails
PM_CALENDAR_STR = "DatetimeNoLeap"
# standard calendar for hindcast experiments
HINDCAST_CALENDAR_STR = "DatetimeProlepticGregorian"

# default kwargs when using concat
# data_vars='minimal' could be added but needs to check that not xr.Dataset
CONCAT_KWARGS = {"coords": "minimal", "compat": "override"}

# name for additional dimension in m2m comparison
M2M_MEMBER_DIM = "forecast_member"

# Valid keywords for aligning inits and verification dates.
VALID_ALIGNMENTS = ["same_inits", "same_init", "same_verifs", "same_verif", "maximize"]

# Valid units for lead dimension
VALID_LEAD_UNITS = [
    "years",
    "seasons",
    "months",
    "weeks",
    "pentads",
    "days",
    "minutes",
    "hours",
    "seconds",
]

# Valid keywords for reference forecast
VALID_REFERENCES = ["uninitialized", "persistence", "climatology"]

# https://github.com/pankajkarman/bias_correction/blob/master/bias_correction.py
BIAS_CORRECTION_BIAS_CORRECTION_METHODS = [
    "modified_quantile",
    "gamma_mapping",
    "basic_quantile",
    "normal_mapping",
]
XCLIM_BIAS_CORRECTION_METHODS = [
    "DetrendedQuantileMapping",
    "LOCI",
    "EmpiricalQuantileMapping",
    # 'ExtremeValues',
    # 'NpdfTransform',
    "PrincipalComponents",
    "QuantileDeltaMapping",
    "Scaling",
]
INTERNAL_BIAS_CORRECTION_METHODS = [
    "additive_mean",
    "multiplicative_mean",
    "multiplicative_std",
]
BIAS_CORRECTION_METHODS = (
    BIAS_CORRECTION_BIAS_CORRECTION_METHODS
    + INTERNAL_BIAS_CORRECTION_METHODS
    + XCLIM_BIAS_CORRECTION_METHODS
)
BIAS_CORRECTION_TRAIN_TEST_SPLIT_METHODS = ["unfair", "unfair-cv", "fair"]
CROSS_VALIDATE_METHODS = ["LOO", False, True]

# seasonality: climpred.set_options(seasonality='...')
GROUPBY_SEASONALITIES = ["dayofyear", "weekofyear", "month", "season"]



import warnings
from copy import deepcopy

import cf_xarray
import numpy as np
import xarray as xr
from dask import is_dask_collection
from IPython.display import display_html
from xarray.core.formatting_html import dataset_repr
from xarray.core.options import OPTIONS as XR_OPTIONS
from xarray.core.utils import Frozen

from .alignment import return_inits_and_verif_dates
from .bias_removal import bias_correction, gaussian_bias_removal, xclim_sdba
from .bootstrap import (
    bootstrap_hindcast,
    bootstrap_perfect_model,
    bootstrap_uninit_pm_ensemble_from_control_cftime,
)
from .checks import (
    _check_valid_reference,
    _check_valud_alignment,
    attach_long_names,
    attach_standard_names,
    has_dataset,
    has_dims,
    has_valid_lead_units,
    is_xarray,
    match_calendars,
    match_initialized_dims,
    match_initialized_vars,
    rename_to_climpred_dims,
)
from .constants import (
    BIAS_CORRECTION_BIAS_CORRECTION_METHODS,
    BIAS_CORRECTION_TRAIN_TEST_SPLIT_METHODS,
    CLIMPRED_DIMS,
    CONCAT_KWARGS,
    CROSS_VALIDATE_METHODS,
    INTERNAL_BIAS_CORRECTION_METHODS,
    M2M_MEMBER_DIM,
    XCLIM_BIAS_CORRECTION_METHODS,
)
from .exceptions import DimensionError, VariableError
from .graphics import plot_ensemble_perfect_model, plot_lead_timeseries_hindcast
from .logging import log_compute_hindcast_header
from .options import OPTIONS
from .prediction import (
    _apply_metric_at_given_lead,
    _get_metric_comparison_dim,
    compute_perfect_model,
)
from .reference import compute_climatology, compute_persistence
from .smoothing import (
    _reset_temporal_axis,
    smooth_goddard_2013,
    spatial_smoothing_xesmf,
    temporal_smoothing,
)
from .utils import (
    broadcast_metric_kwargs_for_rps,
    convert_time_index,
    convert_Timedelta_to_lead_units,
)


def _display_metadata(self):
    """
    This is called in the following case:

    ```
    dp = cp.HindcastEnsemble(dple)
    print(dp)
    ```
    """
    SPACE = "    "
    header = f"<climpred.{type(self).__name__}>"
    summary = header + "\nInitialized Ensemble:\n"
    summary += SPACE + str(self._datasets["initialized"].data_vars)[18:].strip() + "\n"
    if isinstance(self, HindcastEnsemble):
        # Prints out observations and associated variables if they exist.
        # If not, just write "None".
        summary += "Observations:\n"
        if any(self._datasets["observations"]):
            num_obs = len(self._datasets["observations"].data_vars)
            for i in range(1, num_obs + 1):
                summary += (
                    SPACE
                    + str(self._datasets["observations"].data_vars)
                    .split("\n")[i]
                    .strip()
                    + "\n"
                )
        else:
            summary += SPACE + "None\n"
    elif isinstance(self, PerfectModelEnsemble):
        summary += "Control:\n"
        # Prints out control variables if a control is appended. If not,
        # just write "None".
        if any(self._datasets["control"]):
            num_ctrl = len(self._datasets["control"].data_vars)
            for i in range(1, num_ctrl + 1):
                summary += (
                    SPACE
                    + str(self._datasets["control"].data_vars).split("\n")[i].strip()
                    + "\n"
                )
        else:
            summary += SPACE + "None\n"
    if any(self._datasets["uninitialized"]):
        summary += "Uninitialized:\n"
        summary += SPACE + str(self._datasets["uninitialized"].data_vars)[18:].strip()
    else:
        summary += "Uninitialized:\n"
        summary += SPACE + "None"
    return summary


def _display_metadata_html(self):
    header = f"<h4>climpred.{type(self).__name__}</h4>"
    display_html(header, raw=True)
    init_repr_str = dataset_repr(self._datasets["initialized"])
    init_repr_str = init_repr_str.replace("xarray.Dataset", "Initialized Ensemble")
    display_html(init_repr_str, raw=True)

    if isinstance(self, HindcastEnsemble):
        if any(self._datasets["observations"]):
            obs_repr_str = dataset_repr(self._datasets["observations"])
            obs_repr_str = obs_repr_str.replace("xarray.Dataset", "Observations")
            display_html(obs_repr_str, raw=True)
    elif isinstance(self, PerfectModelEnsemble):
        if any(self._datasets["control"]):
            control_repr_str = dataset_repr(self._datasets["control"])
            control_repr_str = control_repr_str.replace(
                "xarray.Dataset", "Control Simulation"
            )
            display_html(control_repr_str, raw=True)

    if any(self._datasets["uninitialized"]):
        uninit_repr_str = dataset_repr(self._datasets["uninitialized"])
        uninit_repr_str = uninit_repr_str.replace("xarray.Dataset", "Uninitialized")
        display_html(uninit_repr_str, raw=True)
    # better would be to aggregate repr_strs and then all return but this fails
    # TypeError: __repr__ returned non-string (type NoneType)
    # workaround return empty string
    return ""


class PredictionEnsemble:
    """
    The main object. This is the super of both `PerfectModelEnsemble` and
    `HindcastEnsemble`. This cannot be called directly by a user, but
    should house functions that both ensemble types can use.
    """

    @is_xarray(1)
    def __init__(self, xobj):
        if isinstance(xobj, xr.DataArray):
            # makes applying prediction functions easier, etc.
            xobj = xobj.to_dataset()
        xobj = rename_to_climpred_dims(xobj)
        has_dims(xobj, ["init", "lead"], "PredictionEnsemble")
        # Check that init is int, cftime, or datetime; convert ints or cftime to
        # datetime.
        xobj = convert_time_index(xobj, "init", "xobj[init]")
        # Put this after `convert_time_index` since it assigns 'years' attribute if the
        # `init` dimension is a `float` or `int`.
        xobj = convert_Timedelta_to_lead_units(xobj)
        has_valid_lead_units(xobj)
        # add metadata
        xobj = attach_standard_names(xobj)
        xobj = attach_long_names(xobj)
        xobj = xobj.cf.add_canonical_attributes(
            verbose=False, override=True, skip="units"
        )
        del xobj.attrs["history"]
        # Add initialized dictionary and reserve sub-dictionary for an uninitialized
        # run.
        self._datasets = {"initialized": xobj, "uninitialized": {}}
        self.kind = "prediction"
        self._temporally_smoothed = None
        self._is_annual_lead = None
        self._warn_if_chunked_along_init_member_time()

    @property
    def coords(self):
        """Dictionary of xarray.DataArray objects corresponding to coordinate
        variables available in all PredictionEnsemble._datasets.
        """
        pe_coords = self.get_initialized().coords.to_dataset()
        for ds in self._datasets.values():
            if isinstance(ds, xr.Dataset):
                pe_coords.update(ds.coords.to_dataset())
        return pe_coords.coords

    @property
    def nbytes(self) -> int:
        """Bytes sizes of all PredictionEnsemble._datasets."""
        return sum(
            [
                sum(v.nbytes for v in ds.variables.values())
                for ds in self._datasets.values()
                if isinstance(ds, xr.Dataset)
            ]
        )

    @property
    def sizes(self):
        """Mapping from dimension names to lengths for all PredictionEnsemble._datasets."""
        pe_dims = dict(self.get_initialized().dims)
        for ds in self._datasets.values():
            if isinstance(ds, xr.Dataset):
                pe_dims.update(dict(ds.dims))
        return pe_dims

    @property
    def dims(self):
        """Mapping from dimension names to lengths all PredictionEnsemble._datasets."""
        return Frozen(self.sizes)

    @property
    def chunks(self):
        """Mapping from chunks all PredictionEnsemble._datasets."""
        pe_chunks = dict(self.get_initialized().chunks)
        for ds in self._datasets.values():
            if isinstance(ds, xr.Dataset):
                for d in ds.chunks:
                    if d not in pe_chunks:
                        pe_chunks.update({d: ds.chunks[d]})
        return Frozen(pe_chunks)

    @property
    def data_vars(self):
        """Dictionary of DataArray objects corresponding to data variables available in all PredictionEnsemble._datasets."""
        varset = set(self.get_initialized().data_vars)
        for ds in self._datasets.values():
            if isinstance(ds, xr.Dataset):
                # take union
                varset = varset & set(ds.data_vars)
        varlist = list(varset)
        return self.get_initialized()[varlist].data_vars

    # when you just print it interactively
    # https://stackoverflow.com/questions/1535327/how-to-print-objects-of-class-using-print
    def __repr__(self):
        if XR_OPTIONS["display_style"] == "html":
            return _display_metadata_html(self)
        else:
            return _display_metadata(self)

    def __len__(self):
        """Number of all variables in all PredictionEnsemble._datasets."""
        return len(self.data_vars)

    def __iter__(self):
        """Iterate over underlying xr.Datasets for initialized, uninitialized, observations."""
        return iter(self._datasets.values())

    def __delitem__(self, key):
        """Remove a variable from this PredictionEnsemble."""
        del self._datasets["initialized"][key]
        for ds in self._datasets.values():
            if isinstance(ds, xr.Dataset):
                if key in ds.data_vars:
                    del ds[key]

    def __contains__(self, key):
        """The 'in' operator will return true or false depending on whether
        'key' is an array in all PredictionEnsemble._datasets or not.
        """
        contained = True
        for ds in self._datasets.values():
            if isinstance(ds, xr.Dataset):
                if key not in ds.data_vars:
                    contained = False
        return contained

    def equals(self, other):
        """Two PredictionEnsembles are equal if they have matching variables and
        coordinates, all of which are equal.
        PredictionEnsembles can still be equal (like pandas objects) if they have NaN
        values in the same locations.
        This method is necessary because `v1 == v2` for ``PredictionEnsembles``
        does element-wise comparisons (like numpy.ndarrays)."""
        if not isinstance(other, PredictionEnsemble):
            return False
        if other.kind != self.kind:
            return False
        equal = True
        try:
            for ds_name in self._datasets.keys():
                if isinstance(self._datasets[ds_name], xr.Dataset):
                    if not self._datasets[ds_name].equals(other._datasets[ds_name]):
                        equal = False
        except Exception:
            return False
        return equal

    def identical(self, other):
        """Like equals, but also checks all dataset attributes and the
        attributes on all variables and coordinates."""
        if not isinstance(other, PredictionEnsemble):
            return False
        if other.kind != self.kind:
            return False
        id = True
        try:
            for ds_name in self._datasets.keys():
                if not self._datasets[ds_name].identical(other._datasets[ds_name]):
                    id = False
        except Exception:
            return False
        return id

    def plot(self, variable=None, ax=None, show_members=False, cmap=None):
        """Plot datasets from PredictionEnsemble.

        Args:
            variable (str or None): `variable` to show. Defaults to first in data_vars.
            ax (plt.axes): Axis to use in plotting. By default, creates a new axis.
            show_members (bool): whether to display all members individually.
                Defaults to False.
            cmap (str): Name of matplotlib-recognized colorbar. Defaults to `jet` for
                `HindcastEnsemble` and `tab10` for `PerfectModelEnsemble`.

        Returns:
            ax: plt.axes

        """
        if self.kind == "hindcast":
            if cmap is None:
                cmap = "jet"
            return plot_lead_timeseries_hindcast(
                self, variable=variable, ax=ax, show_members=show_members, cmap=cmap
            )
        elif self.kind == "perfect":
            if cmap is None:
                cmap = "tab10"
            return plot_ensemble_perfect_model(
                self, variable=variable, ax=ax, show_members=show_members, cmap=cmap
            )

    def _math(self, other, operator):
        """Helper function for __add__, __sub__, __mul__, __truediv__.

        Allows math operations with type:
            - int
            - float
            - np.ndarray
            - xr.DataArray without new dimensions
            - xr.Dataset without new dimensions or variables

        """
        assert isinstance(operator, str)

        def add(a, b):
            return a + b

        def sub(a, b):
            return a - b

        def mul(a, b):
            return a * b

        def div(a, b):
            return a / b

        ALLOWED_TYPES_FOR_MATH_OPERATORS = [
            int,
            float,
            np.ndarray,
            xr.DataArray,
            xr.Dataset,
            type(self),
        ]
        OPERATOR_STR = {
            "add": "+",
            "sub": "-",
            "mul": "*",
            "div": "/",
        }
        error_str = f"Cannot use {type(self)} {OPERATOR_STR[operator]} {type(other)}"

        # catch undefined types for other
        if not isinstance(other, tuple(ALLOWED_TYPES_FOR_MATH_OPERATORS)):
            raise TypeError(
                f"{error_str} because type {type(other)} not supported. "
                f"Please choose from {ALLOWED_TYPES_FOR_MATH_OPERATORS}."
            )
        # catch other dimensions in other
        if isinstance(other, tuple([xr.Dataset, xr.DataArray])):
            if not set(other.dims).issubset(self._datasets["initialized"].dims):
                raise DimensionError(f"{error_str} containing new dimensions.")
        # catch xr.Dataset with different data_vars
        if isinstance(other, xr.Dataset):
            if list(other.data_vars) != list(self._datasets["initialized"].data_vars):
                raise VariableError(
                    f"{error_str} with new `data_vars`. Please use {type(self)} "
                    f"{operator} {type(other)} only with same `data_vars`. Found "
                    f"initialized.data_vars = "
                    f' {list(self._datasets["initialized"].data_vars)} vs. '
                    f"other.data_vars = {list(other.data_vars)}."
                )

        operator = eval(operator)

        if isinstance(other, PredictionEnsemble):
            # Create temporary copy to modify to avoid inplace operation.
            datasets = self._datasets.copy()
            for dataset in datasets:
                # Some pre-allocated entries might be empty, such as 'uninitialized'
                if isinstance(other._datasets[dataset], xr.Dataset) and isinstance(
                    self._datasets[dataset], xr.Dataset
                ):
                    datasets[dataset] = operator(
                        datasets[dataset], other._datasets[dataset]
                    )
            return self._construct_direct(datasets, kind=self.kind)
        else:
            return self._apply_func(operator, other)

    def __add__(self, other):
        return self._math(other, operator="add")

    def __sub__(self, other):
        return self._math(other, operator="sub")

    def __mul__(self, other):
        return self._math(other, operator="mul")

    def __truediv__(self, other):
        return self._math(other, operator="div")

    def __getitem__(self, varlist):
        """Allows subsetting data variable from PredictionEnsemble as from xr.Dataset.

        Args:
            * varlist (list of str, str): list of names or name of data variable(s) to
                subselect
        """
        if isinstance(varlist, str):
            varlist = [varlist]
        if not isinstance(varlist, list):
            raise ValueError(
                "Please subset PredictionEnsemble as you would subset an xr.Dataset "
                "with a list or single string of variable name(s), found "
                f"{type(varlist)}."
            )

        def sel_vars(ds, varlist):
            return ds[varlist]

        return self._apply_func(sel_vars, varlist)

    def __getattr__(self, name):
        """Allows for xarray methods to be applied to our prediction objects.

        Args:
            * name: Function, e.g., .isel() or .sum().
        """

        def wrapper(*args, **kwargs):
            """Applies arbitrary function to all datasets in the PredictionEnsemble
            object.

            Got this from: https://stackoverflow.com/questions/41919499/
            how-to-call-undefined-methods-sequentially-in-python-class
            """

            def _apply_xr_func(v, name, *args, **kwargs):
                """Handles exceptions in our dictionary comprehension.

                In other words, this will skip applying the arbitrary function
                to a sub-dataset if a ValueError is thrown. This specifically
                targets cases where certain datasets don't have the given
                dim that's being called. E.g., ``.isel(lead=0)`` should only
                be applied to the initialized dataset.

                Reference:
                  * https://stackoverflow.com/questions/1528237/
                    how-to-handle-exceptions-in-a-list-comprehensions
                """
                try:
                    return getattr(v, name)(*args, **kwargs)
                # ValueError : Cases such as .sum(dim='time'). This doesn't apply
                #              it to the given dataset if the dimension doesn't exist.
                # KeyError : Cases where a function calls the index of a Dataset. Such
                #            as ds[dim] and the dim doesn't exist as a key.
                # DimensionError: This accounts for our custom error when applying
                # some stats functions.
                except (ValueError, KeyError, DimensionError) as e:
                    if args == tuple():
                        func_name = False
                    else:
                        if callable(args[0]):
                            func_name = args[0].__name__
                        else:  # for xarray calls like pe.mean()
                            func_name = False
                    dim = kwargs.get("dim", False)
                    error_type = type(e).__name__
                    if func_name:
                        if len(args) > 1:
                            msg = f"{func_name}({args[1:]}, {kwargs}) failed\n{error_type}: {e}"
                        else:
                            msg = f"{func_name}({kwargs}) failed\n{error_type}: {e}"
                    else:
                        msg = f"xr.{name}({args}, {kwargs}) failed\n{error_type}: {e}"
                    if set(["lead", "init"]).issubset(set(v.dims)):  # initialized
                        if dim not in v.dims:
                            if OPTIONS["warn_for_failed_PredictionEnsemble_xr_call"]:
                                warnings.warn(f"Error due to initialized:  {msg}")
                    elif set(["time"]).issubset(
                        set(v.dims)
                    ):  # uninitialized, control, verification
                        if dim not in v.dims:
                            if OPTIONS["warn_for_failed_PredictionEnsemble_xr_call"]:
                                warnings.warn(
                                    f"Error due to verification/control/uninitialized: {msg}"
                                )
                    else:
                        if OPTIONS["warn_for_failed_PredictionEnsemble_xr_call"]:
                            warnings.warn(msg)
                    return v

            return self._apply_func(_apply_xr_func, name, *args, **kwargs)

        return wrapper

    @classmethod
    def _construct_direct(cls, datasets, kind):
        """Shortcut around __init__ for internal use to avoid inplace
        operations.

        Pulled from xarrray Dataset class.
        https://github.com/pydata/xarray/blob/master/xarray/core/dataset.py
        """
        obj = object.__new__(cls)
        obj._datasets = datasets
        obj.kind = kind
        obj._warn_if_chunked_along_init_member_time()
        return obj

    def _apply_func(self, func, *args, **kwargs):
        """Apply a function to all datasets in a `PredictionEnsemble`."""
        # Create temporary copy to modify to avoid inplace operation.
        datasets = self._datasets.copy()

        # More explicit than nested dictionary comprehension.
        for key, ds in datasets.items():
            # If ds is xr.Dataset, apply the function directly to it. else, e.g. for {} ignore
            if isinstance(ds, xr.Dataset):
                dim = kwargs.get("dim", "")
                if "_or_" in dim:
                    dims = dim.split("_or_")
                    if set(dims).issubset(ds.dims):
                        raise ValueError(
                            f"{dims} cannot be both in {key} dataset, found {ds.dims}"
                        )
                    kwargs_dim0 = kwargs.copy()
                    kwargs_dim0["dim"] = dims[0]
                    kwargs_dim1 = kwargs.copy()
                    kwargs_dim1["dim"] = dims[1]
                    if dims[0] in ds.dims and dims[1] not in ds.dims:
                        datasets.update({key: func(ds, *args, **kwargs_dim0)})
                    if dims[1] in ds.dims and dims[0] not in ds.dims:
                        datasets.update({key: func(ds, *args, **kwargs_dim1)})
                else:
                    datasets.update({key: func(ds, *args, **kwargs)})
        # Instantiates new object with the modified datasets.
        return self._construct_direct(datasets, kind=self.kind)

    def get_initialized(self):
        """Returns the xarray dataset for the initialized ensemble."""
        return self._datasets["initialized"]

    def get_uninitialized(self):
        """Returns the xarray dataset for the uninitialized ensemble."""
        return self._datasets["uninitialized"]

    def smooth(self, smooth_kws=None, how="mean", **xesmf_kwargs):
        """Smooth all entries of PredictionEnsemble in the same manner to be
        able to still calculate prediction skill afterwards.

        Args:
            smooth_kws (dict or str): Dictionary to specify the dims to
                smooth compatible with
                :py:func:`~climpred.smoothing.spatial_smoothing_xesmf` or
                :py:func:`~climpred.smoothing.temporal_smoothing`.
                Shortcut for Goddard et al. 2013 recommendations:
                'goddard2013'. Defaults to None.
            how (str): how to smooth temporally. From ['mean','sum']. Defaults to
                'mean'.
            **xesmf_kwargs (args): kwargs passed to
                :py:func:`~climpred.smoothing.spatial_smoothing_xesmf`

        Examples:
            >>> PerfectModelEnsemble.get_initialized().lead.size
            20
            >>> PerfectModelEnsemble.smooth({'lead':4}, how='sum').get_initialized().lead.size
            17

            >>> HindcastEnsemble_3D.smooth({'lon':1, 'lat':1})
            <climpred.HindcastEnsemble>
            Initialized Ensemble:
                SST      (init, lead, lat, lon) float64 -0.3236 -0.3161 -0.3083 ... 0.0 0.0
            Observations:
                SST      (time, lat, lon) float64 0.002937 0.001561 0.002587 ... 0.0 0.0 0.0
            Uninitialized:
                None

            ``smooth`` simultaneously aggregates spatially listening to ``lon`` and ``lat`` and temporally listening to ``lead`` or ``time``.

            >>> HindcastEnsemble_3D.smooth({'lead': 2, 'lat': 5, 'lon': 4}).get_initialized().coords
            Coordinates:
              * init     (init) object 1954-01-01 00:00:00 ... 2017-01-01 00:00:00
              * lead     (lead) int32 1 2 3 4 5 6 7 8 9
              * lon      (lon) float64 250.8 254.8 258.8 262.8
              * lat      (lat) float64 -9.75 -4.75
            >>> HindcastEnsemble_3D.smooth('goddard2013').get_initialized().coords
            Coordinates:
              * init     (init) object 1954-01-01 00:00:00 ... 2017-01-01 00:00:00
              * lead     (lead) int32 1 2 3 4 5 6 7
              * lon      (lon) float64 250.8 255.8 260.8 265.8
              * lat      (lat) float64 -9.75 -4.75


        """
        if not smooth_kws:
            return self
        # get proper smoothing function based on smooth args
        if isinstance(smooth_kws, str):
            if "goddard" in smooth_kws:
                if self._is_annual_lead:
                    smooth_fct = smooth_goddard_2013
                    tsmooth_kws = {"lead": 4}  # default
                    d_lon_lat_kws = {"lon": 5, "lat": 5}  # default
                else:
                    raise ValueError(
                        "`goddard2013` smoothing only available for annual leads."
                    )
            else:
                raise ValueError(
                    'Please provide from list of available smoothings: \
                     ["goddard2013"]'
                )
        # TODO: actively searches for lot and lat in dims. Maybe this part of the code
        # could be more robust in how it finds these two spatial dimensions regardless
        # of name. Optional work in progress comment.
        elif isinstance(smooth_kws, dict):
            non_time_dims = [
                dim for dim in smooth_kws.keys() if dim not in ["time", "lead"]
            ]
            if len(non_time_dims) > 0:
                non_time_dims = non_time_dims[0]
            # goddard when time_dim and lon/lat given
            if ("lon" in smooth_kws or "lat" in smooth_kws) and (
                "lead" in smooth_kws or "time" in smooth_kws
            ):
                smooth_fct = smooth_goddard_2013
                # separate lon, lat keywords into d_lon_lat_kws
                d_lon_lat_kws = dict()
                tsmooth_kws = dict()
                for c in ["lon", "lat"]:
                    if c in smooth_kws:
                        d_lon_lat_kws[c] = smooth_kws[c]
                for c in ["lead", "time"]:
                    if c in smooth_kws:
                        tsmooth_kws[c] = smooth_kws[c]
            # else only one smoothing operation
            elif "lon" in smooth_kws or "lat" in smooth_kws:
                smooth_fct = spatial_smoothing_xesmf
                d_lon_lat_kws = smooth_kws
                tsmooth_kws = None
            elif "lead" in smooth_kws or "time" in smooth_kws:
                smooth_fct = temporal_smoothing
                d_lon_lat_kws = None
                tsmooth_kws = smooth_kws
            else:
                raise ValueError(
                    'Please provide kwargs to fulfill functions: \
                     ["spatial_smoothing_xesmf", "temporal_smoothing"].'
                )
        else:
            raise ValueError(
                "Please provide kwargs as dict or str and not", type(smooth_kws)
            )
        self = self.map(
            smooth_fct,
            tsmooth_kws=tsmooth_kws,
            d_lon_lat_kws=d_lon_lat_kws,
            how=how,
            **xesmf_kwargs,
        )
        if smooth_fct == smooth_goddard_2013 or smooth_fct == temporal_smoothing:
            self._temporally_smoothed = tsmooth_kws
        return self

    def _warn_if_chunked_along_init_member_time(self):
        """Warn upon instantiation when CLIMPRED_DIMS except ``lead`` are chunked with
        more than one chunk to show how to circumvent ``xskillscore`` chunking
        ``ValueError``."""
        suggest_one_chunk = []
        for d in self.chunks:
            if d in ["time", "init", "member"]:
                if len(self.chunks[d]) > 1:
                    suggest_one_chunk.append(d)
        if len(suggest_one_chunk) > 0:
            name = (
                str(type(self))
                .replace("<class 'climpred.classes.", "")
                .replace("'>", "")
            )
            # init cannot be dim when time chunked
            suggest_one_chunk_time_to_init = suggest_one_chunk.copy()
            if "time" in suggest_one_chunk_time_to_init:
                suggest_one_chunk_time_to_init.remove("time")
                suggest_one_chunk_time_to_init.append("init")
            msg = f"{name} is chunked along dimensions {suggest_one_chunk} with more than one chunk. `{name}.chunks={self.chunks}`.\nYou cannot call `{name}.verify` or `{name}.bootstrap` in combination with any of {suggest_one_chunk_time_to_init} passed as `dim`. In order to do so, please rechunk {suggest_one_chunk} with `{name}.chunk({{dim:-1}}).verify(dim=dim).`\nIf you do not want to use dimensions {suggest_one_chunk_time_to_init} in `{name}.verify(dim=dim)`, you can disregard this warning."
            # chunk lead:1 in HindcastEnsemble
            if self.kind == "hindcast":
                msg += '\nIn `HindcastEnsemble`s you may also create one chunk per lead, as the `climpred` internally loops over lead, e.g. `.chunk({{"lead": 1}}).verify().`'
            # chunk auto on non-climpred dims
            ndims = list(self.sizes)
            for d in CLIMPRED_DIMS:
                if d in ndims:
                    ndims.remove(d)
            if len(ndims) > 0:
                msg += f'\nConsider chunking embarassingly parallel dimensions such as {ndims} automatically, i.e. `{name}.chunk({ndims[0]}="auto").verify(...).'
            warnings.warn(msg)


class PerfectModelEnsemble(PredictionEnsemble):
    """An object for "perfect model" climate prediction ensembles.

    `PerfectModelEnsemble` is a sub-class of `PredictionEnsemble`. It tracks
    the control run used to initialize the ensemble for easy computations,
    bootstrapping, etc.

    This object is built on `xarray` and thus requires the input object to
    be an `xarray` Dataset or DataArray.
    """

    def __init__(self, xobj):
        """Create a `PerfectModelEnsemble` object by inputting output from the
        control run in `xarray` format.

        Args:
          xobj (xarray object):
            decadal prediction ensemble output.

        Attributes:
            control: Dictionary of control run associated with the initialized
                     ensemble.
            uninitialized: Dictionary of uninitialized run that is
                           bootstrapped from the initialized run.
        """

        super().__init__(xobj)
        # Reserve sub-dictionary for the control simulation.
        self._datasets.update({"control": {}})
        self.kind = "perfect"

    def _apply_climpred_function(self, func, input_dict=None, **kwargs):
        """Helper function to loop through observations and apply an arbitrary climpred
        function.

        Args:
            func (function): climpred function to apply to object.
            input_dict (dict): dictionary with the following things:
                * ensemble: initialized or uninitialized ensemble.
                * control: control dictionary from HindcastEnsemble.
                * init (bool): True if the initialized ensemble, False if uninitialized.
        """
        ensemble = input_dict["ensemble"]
        control = input_dict["control"]
        init = input_dict["init"]
        init_vars, ctrl_vars = self._vars_to_drop(init=init)
        ensemble = ensemble.drop_vars(init_vars)
        if control:
            control = control.drop_vars(ctrl_vars)
        return func(ensemble, control, **kwargs)

    def _vars_to_drop(self, init=True):
        """Returns list of variables to drop when comparing
        initialized/uninitialized to a control.

        This is useful if the two products being compared do not share the same
        variables. I.e., if the control has ['SST'] and the initialized has
        ['SST', 'SALT'], this will return a list with ['SALT'] to be dropped
        from the initialized.

        Args:
          init (bool, default True):
            If `True`, check variables on the initialized.
            If `False`, check variables on the uninitialized.

        Returns:
          Lists of variables to drop from the initialized/uninitialized
          and control Datasets.
        """
        init_str = "initialized" if init else "uninitialized"
        init_vars = list(self._datasets[init_str])
        # only drop if control present
        if self._datasets["control"]:
            ctrl_vars = list(self._datasets["control"])
            # Make lists of variables to drop that aren't in common
            # with one another.
            init_vars_to_drop = list(set(init_vars) - set(ctrl_vars))
            ctrl_vars_to_drop = list(set(ctrl_vars) - set(init_vars))
        else:
            init_vars_to_drop, ctrl_vars_to_drop = [], []
        return init_vars_to_drop, ctrl_vars_to_drop

    @is_xarray(1)
    def add_control(self, xobj):
        """Add the control run that initialized the climate prediction
        ensemble.

        Args:
            xobj (xarray object): Dataset/DataArray of the control run.
        """
        # NOTE: These should all be decorators.
        if isinstance(xobj, xr.DataArray):
            xobj = xobj.to_dataset()
        match_initialized_dims(self._datasets["initialized"], xobj)
        match_initialized_vars(self._datasets["initialized"], xobj)
        # Check that init is int, cftime, or datetime; convert ints or cftime to
        # datetime.
        xobj = convert_time_index(xobj, "time", "xobj[init]")
        # Check that converted/original cftime calendar is the same as the
        # initialized calendar to avoid any alignment errors.
        match_calendars(self._datasets["initialized"], xobj, kind2="control")
        datasets = self._datasets.copy()
        datasets.update({"control": xobj})
        return self._construct_direct(datasets, kind="perfect")

    def generate_uninitialized(self):
        """Generate an uninitialized ensemble by bootstrapping the
        initialized prediction ensemble.

        Returns:
            Bootstrapped (uninitialized) ensemble as a Dataset.
        """
        has_dataset(
            self._datasets["control"], "control", "generate an uninitialized ensemble."
        )

        uninit = bootstrap_uninit_pm_ensemble_from_control_cftime(
            self._datasets["initialized"], self._datasets["control"]
        )
        datasets = self._datasets.copy()
        datasets.update({"uninitialized": uninit})
        return self._construct_direct(datasets, kind="perfect")

    def get_control(self):
        """Returns the control as an xarray dataset."""
        return self._datasets["control"]

    def verify(
        self,
        metric=None,
        comparison=None,
        dim=None,
        reference=None,
        **metric_kwargs,
    ):
        """Verify initialized predictions against a configuration of other ensemble members.

        .. note::
            The configuration of the other ensemble members is based off of the
            ``comparison`` keyword argument.

        Args:
            metric (str, :py:class:`~climpred.metrics.Metric`): Metric to apply in the
                comparison. See `metrics </metrics.html>`_.
            comparison (str, :py:class:`~climpred.comparisons.Comparison`): How to
                compare the initialized prediction ensemble with itself, see
                `comparisons </comparisons.html>`_.
            dim (str, list of str): Dimension(s) over which to apply ``metric``.
                ``dim`` is passed on to xskillscore.{metric} and includes xskillscore's
                ``member_dim``. ``dim`` should contain ``member`` when ``comparison``
                is probabilistic but should not contain ``member`` when
                ``comparison=e2c``. Defaults to ``None`` meaning that all dimensions
                other than ``lead`` are reduced.
            reference (str, list of str): Type of reference forecasts with which to
                verify. One or more of ['uninitialized', 'persistence', 'climatology'].
            **metric_kwargs (optional): Arguments passed to ``metric``.

        Returns:
            Dataset with dimension skill reduced by dim containing initialized and
            reference skill(s) if specified.

        Example:
            Root mean square error (``rmse``) comparing every member with the
            ensemble mean forecast (``m2e``) for all leads reducing dimensions
            ``init`` and ``member``:

            >>> PerfectModelEnsemble.verify(metric='rmse', comparison='m2e',
            ...     dim=['init','member'])
            <xarray.Dataset>
            Dimensions:  (lead: 20)
            Coordinates:
              * lead     (lead) int64 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 19 20
            Data variables:
                tos      (lead) float32 0.1028 0.1249 0.1443 0.1707 ... 0.2113 0.2452 0.2297


            Pearson's Anomaly Correlation ('acc') comparing every member to every
            other member (``m2m``) reducing dimensions ``member`` and ``init`` while
            also calculating reference skill for the ``persistence``, ``climatology``
            and ``uninitialized`` forecast.

            >>> PerfectModelEnsemble.verify(metric='acc', comparison='m2m',
            ...     dim=['init', 'member'],
            ...     reference=['persistence', 'climatology' ,'uninitialized'])
            <xarray.Dataset>
            Dimensions:  (skill: 4, lead: 20)
            Coordinates:
              * lead     (lead) int64 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 19 20
              * skill    (skill) <U13 'initialized' 'persistence' ... 'uninitialized'
            Data variables:
                tos      (skill, lead) float64 0.7941 0.7489 0.5623 ... 0.1327 0.4547 0.3253
        """
        reference = _check_valid_reference(reference)
        input_dict = {
            "ensemble": self._datasets["initialized"],
            "control": self._datasets["control"]
            if isinstance(self._datasets["control"], xr.Dataset)
            else None,
            "init": True,
        }
        result = self._apply_climpred_function(
            compute_perfect_model,
            input_dict=input_dict,
            metric=metric,
            comparison=comparison,
            dim=dim,
            add_attrs=False,
            **metric_kwargs,
        )
        if self._temporally_smoothed:
            result = _reset_temporal_axis(result, self._temporally_smoothed, dim="lead")
            result["lead"].attrs = self.get_initialized().lead.attrs
        # compute reference skills
        if reference:
            for r in reference:
                dim_orig = deepcopy(dim)  # preserve dim, because
                ref_compute_kwargs = metric_kwargs.copy()  # persistence changes dim
                ref_compute_kwargs.update({"dim": dim_orig, "metric": metric})
                if r != "persistence":
                    ref_compute_kwargs["comparison"] = comparison
                ref = getattr(self, f"_compute_{r}")(**ref_compute_kwargs)
                result = xr.concat([result, ref], dim="skill", **CONCAT_KWARGS)
            result = result.assign_coords(skill=["initialized"] + reference)
        return result.squeeze()

    def _compute_uninitialized(
        self, metric=None, comparison=None, dim=None, **metric_kwargs
    ):
        """Verify the bootstrapped uninitialized run against itself.

        .. note::
            The configuration of the other ensemble members is based off of the
            ``comparison`` keyword argument.

        Args:
            metric (str, :py:class:`~climpred.metrics.Metric`): Metric to apply in the
                comparison. See `metrics </metrics.html>`_.
            comparison (str, :py:class:`~climpred.comparisons.Comparison`): How to
                compare the uninitialized against itself, see
                `comparisons </comparisons.html>`_.
            dim (str, list of str): Dimension(s) over which to apply metric.
                ``dim`` is passed on to xskillscore.{metric} and includes xskillscore's
                ``member_dim``. ``dim`` should contain ``member`` when ``comparison``
                is probabilistic but should not contain ``member`` when
                ``comparison=e2c``. Defaults to ``None``, meaning that all dimensions
                other than ``lead`` are reduced.
            **metric_kwargs (optional): Arguments passed to ``metric``.

        Returns:
            Dataset with dimension skill containing initialized and reference skill(s).
        """
        has_dataset(
            self._datasets["uninitialized"],
            "uninitialized",
            "compute an uninitialized metric",
        )
        input_dict = {
            "ensemble": self._datasets["uninitialized"],
            "control": self._datasets["control"]
            if isinstance(self._datasets["control"], xr.Dataset)
            else None,
            "init": False,
        }
        if dim is None:
            dim = list(self._datasets["initialized"].isel(lead=0).dims)
        res = self._apply_climpred_function(
            compute_perfect_model,
            input_dict=input_dict,
            metric=metric,
            comparison=comparison,
            dim=dim,
            **metric_kwargs,
        )
        if self._temporally_smoothed:
            res = _reset_temporal_axis(res, self._temporally_smoothed, dim="lead")
            res["lead"].attrs = self.get_initialized().lead.attrs
        return res

    def _compute_persistence(self, metric=None, dim=None, **metric_kwargs):
        """Verify a simple persistence forecast of the control run against itself.

        Args:
            metric (str, :py:class:`~climpred.metrics.Metric`): Metric to use when
            verifying skill of the persistence forecast. See `metrics </metrics.html>`_.
            dim (str, list of str): Dimension(s) over which to apply metric.
                ``dim`` is passed on to xskillscore.{metric} and includes xskillscore's
                ``member_dim``. ``dim`` should contain ``member`` when ``comparison``
                is probabilistic but should not contain ``member`` when
                ``comparison=e2c``. Defaults to ``None``, meaning that all dimensions
                other than ``lead`` are reduced.
            **metric_kwargs (optional): Arguments passed to ``metric``.

        Returns:
            Dataset of persistence forecast results.

        Reference:
            * Chapter 8 (Short-Term Climate Prediction) in
              Van den Dool, Huug. Empirical methods in short-term climate
              prediction. Oxford University Press, 2007.
        """
        has_dataset(
            self._datasets["control"], "control", "compute a persistence forecast"
        )
        input_dict = {
            "ensemble": self._datasets["initialized"],
            "control": self._datasets["control"],
            "init": True,
        }
        if dim is None:
            dim = list(self._datasets["initialized"].isel(lead=0).dims)
        res = self._apply_climpred_function(
            compute_persistence,
            input_dict=input_dict,
            metric=metric,
            alignment="same_inits",
            dim=dim,
            **metric_kwargs,
        )
        if self._temporally_smoothed:
            res = _reset_temporal_axis(res, self._temporally_smoothed, dim="lead")
            res["lead"].attrs = self.get_initialized().lead.attrs
        return res

    def _compute_climatology(
        self, metric=None, comparison=None, dim=None, **metric_kwargs
    ):
        """Verify a climatology forecast of the control run against itself.

        Args:
            metric (str, :py:class:`~climpred.metrics.Metric`): Metric to use when
            verifying skill of the persistence forecast. See `metrics </metrics.html>`_.
            dim (str, list of str): Dimension(s) over which to apply metric.
                ``dim`` is passed on to xskillscore.{metric} and includes xskillscore's
                ``member_dim``. ``dim`` should contain ``member`` when ``comparison``
                is probabilistic but should not contain ``member`` when
                ``comparison=e2c``. Defaults to ``None``, meaning that all dimensions
                other than ``lead`` are reduced.
            **metric_kwargs (optional): Arguments passed to ``metric``.

        Returns:
            Dataset of persistence forecast results.

        Reference:
            * Chapter 8 (Short-Term Climate Prediction) in
              Van den Dool, Huug. Empirical methods in short-term climate
              prediction. Oxford University Press, 2007.
        """
        input_dict = {
            "ensemble": self._datasets["initialized"],
            "control": self._datasets["control"]
            if isinstance(self._datasets["control"], xr.Dataset)
            else None,
            "init": True,
        }
        if dim is None:
            dim = list(self.get_initialized().isel(lead=0).dims)
        res = self._apply_climpred_function(
            compute_climatology,
            input_dict=input_dict,
            metric=metric,
            comparison=comparison,
            dim=dim,
            **metric_kwargs,
        )
        if self._temporally_smoothed:
            res = _reset_temporal_axis(res, self._temporally_smoothed, dim="lead")
            res["lead"].attrs = self.get_initialized().lead.attrs
        return res

    def bootstrap(
        self,
        metric=None,
        comparison=None,
        dim=None,
        reference=None,
        iterations=None,
        sig=95,
        pers_sig=None,
        **metric_kwargs,
    ):
        """Bootstrap with replacement according to Goddard et al. 2013.

        Args:
            metric (str, :py:class:`~climpred.metrics.Metric`): Metric to verify
                bootstrapped skill, see `metrics </metrics.html>`_.
            comparison (str, :py:class:`~climpred.comparisons.Comparison`): Comparison
                passed to verify, see `comparisons </comparisons.html>`_.
            dim (str, list of str): Dimension(s) over which to apply metric.
                ``dim`` is passed on to xskillscore.{metric} and includes xskillscore's
                ``member_dim``. ``dim`` should contain ``member`` when ``comparison``
                is probabilistic but should not contain ``member`` when
                ``comparison=e2c``. Defaults to ``None`` meaning that all dimensions
                other than ``lead`` are reduced.
            reference (str, list of str): Type of reference forecasts with which to
                verify. One or more of ['uninitialized', 'persistence', 'climatology'].
                If None or empty, returns no p value.
            iterations (int): Number of resampling iterations for bootstrapping with
                replacement. Recommended >= 500.
            sig (int, default 95): Significance level in percent for deciding whether
                uninitialized and persistence beat initialized skill.
            pers_sig (int): If not ``None``, the separate significance level for
                persistence. Defaults to ``None``, or the same significance as ``sig``.
            **metric_kwargs (optional): arguments passed to ``metric``.

        Returns:
            xr.Datasets: with dimensions ``results`` (holding ``verify skill``, ``p``,
            ``low_ci`` and ``high_ci``) and ``skill`` (holding ``initialized``,
            ``persistence`` and/or ``uninitialized``):
                * results='verify skill', skill='initialized':
                    mean initialized skill
                * results='high_ci', skill='initialized':
                    high confidence interval boundary for initialized skill
                * results='p', skill='uninitialized':
                    p value of the hypothesis that the
                    difference of skill between the initialized and
                    uninitialized simulations is smaller or equal to zero
                    based on bootstrapping with replacement.
                * results='p', skill='persistence':
                    p value of the hypothesis that the
                    difference of skill between the initialized and persistenceistence
                    simulations is smaller or equal to zero based on
                    bootstrapping with replacement.

        Reference:
            * Goddard, L., A. Kumar, A. Solomon, D. Smith, G. Boer, P.
              Gonzalez, V. Kharin, et al. “A Verification Framework for
              Interannual-to-Decadal Predictions Experiments.” Climate
              Dynamics 40, no. 1–2 (January 1, 2013): 245–72.
              https://doi.org/10/f4jjvf.

        Example:
            Calculate the Pearson's Anomaly Correlation ('acc') comparing every member
            to every other member (``m2m``) reducing dimensions ``member`` and
            ``init`` 50 times after resampling ``member`` dimension with replacement.
            Also calculate reference skill for the ``persistence``, ``climatology``
            and ``uninitialized`` forecast and compare whether initialized skill is
            better than reference skill: Returns verify skill, probability that
            reference forecast performs better than initialized and the lower and
            upper bound of the resample.

            >>> PerfectModelEnsemble.bootstrap(metric='acc', comparison='m2m',
            ...     dim=['init', 'member'], iterations=50, resample_dim='member',
            ...     reference=['persistence', 'climatology' ,'uninitialized'])
            <xarray.Dataset>
            Dimensions:  (skill: 4, results: 4, lead: 20)
            Coordinates:
              * lead     (lead) int64 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 19 20
              * results  (results) <U12 'verify skill' 'p' 'low_ci' 'high_ci'
              * skill    (skill) <U13 'initialized' 'persistence' ... 'uninitialized'
            Data variables:
                tos      (skill, results, lead) float64 0.7941 0.7489 ... 0.1494 0.1466
            Attributes:
                prediction_skill:            calculated by climpred https://climpred.read...
                number_of_initializations:   12
                number_of_members:           10
                alignment:                   same_verifs
                metric:                      pearson_r
                comparison:                  m2m
                dim:                         ['init', 'member']
                units:                       None
                confidence_interval_levels:  0.975-0.025
                bootstrap_iterations:        50
                p:                           probability that reference performs better t...

        """
        if iterations is None:
            raise ValueError("Designate number of bootstrapping `iterations`.")
        reference = _check_valid_reference(reference)
        has_dataset(self._datasets["control"], "control", "iteration")
        input_dict = {
            "ensemble": self._datasets["initialized"],
            "control": self._datasets["control"],
            "init": True,
        }
        return self._apply_climpred_function(
            bootstrap_perfect_model,
            input_dict=input_dict,
            metric=metric,
            comparison=comparison,
            dim=dim,
            reference=reference,
            sig=sig,
            iterations=iterations,
            pers_sig=pers_sig,
            **metric_kwargs,
        )


class HindcastEnsemble(PredictionEnsemble):
    """An object for climate prediction ensembles initialized by a data-like
    product.

    `HindcastEnsemble` is a sub-class of `PredictionEnsemble`. It tracks a single
    verification dataset (i.e., observations) associated with the hindcast ensemble
    for easy computation across multiple variables.

    This object is built on `xarray` and thus requires the input object to
    be an `xarray` Dataset or DataArray.
    """

    def __init__(self, xobj):
        """Create a `HindcastEnsemble` object by inputting output from a
        prediction ensemble in `xarray` format.

        Args:
          xobj (xarray object):
            decadal prediction ensemble output.

        Attributes:
          observations: Dictionary of verification data to associate with the decadal
              prediction ensemble.
          uninitialized: Dictionary of companion (or bootstrapped)
              uninitialized ensemble run.
        """
        super().__init__(xobj)
        self._datasets.update({"observations": {}})
        self.kind = "hindcast"

    def _apply_climpred_function(self, func, init, **kwargs):
        """Helper function to loop through verification data and apply an arbitrary
        climpred function.

        Args:
            func (function): climpred function to apply to object.
            init (bool): Whether or not it's the initialized ensemble.
        """
        hind = self._datasets["initialized"]
        verif = self._datasets["observations"]
        drop_init, drop_obs = self._vars_to_drop(init=init)
        return func(hind.drop_vars(drop_init), verif.drop_vars(drop_obs), **kwargs)

    def _vars_to_drop(self, init=True):
        """Returns list of variables to drop when comparing
        initialized/uninitialized to observations.

        This is useful if the two products being compared do not share the same
        variables. I.e., if the observations have ['SST'] and the initialized has
        ['SST', 'SALT'], this will return a list with ['SALT'] to be dropped
        from the initialized.

        Args:
          init (bool, default True):
            If ``True``, check variables on the initialized.
            If ``False``, check variables on the uninitialized.

        Returns:
          Lists of variables to drop from the initialized/uninitialized
          and observational Datasets.
        """
        if init:
            init_vars = [var for var in self._datasets["initialized"].data_vars]
        else:
            init_vars = [var for var in self._datasets["uninitialized"].data_vars]
        obs_vars = [var for var in self._datasets["observations"].data_vars]
        # Make lists of variables to drop that aren't in common
        # with one another.
        init_vars_to_drop = list(set(init_vars) - set(obs_vars))
        obs_vars_to_drop = list(set(obs_vars) - set(init_vars))
        return init_vars_to_drop, obs_vars_to_drop

    @is_xarray(1)
    def add_observations(self, xobj):
        """Add verification data against which to verify the initialized ensemble.

        Args:
            xobj (xarray object): Dataset/DataArray to append to the
                ``HindcastEnsemble`` object.
        """
        if isinstance(xobj, xr.DataArray):
            xobj = xobj.to_dataset()
        match_initialized_dims(self._datasets["initialized"], xobj)
        match_initialized_vars(self._datasets["initialized"], xobj)
        # Check that time is int, cftime, or datetime; convert ints or cftime to
        # datetime.
        xobj = convert_time_index(xobj, "time", "xobj[init]")
        # Check that converted/original cftime calendar is the same as the
        # initialized calendar to avoid any alignment errors.
        match_calendars(self._datasets["initialized"], xobj)
        datasets = self._datasets.copy()
        datasets.update({"observations": xobj})
        return self._construct_direct(datasets, kind="hindcast")

    @is_xarray(1)
    def add_uninitialized(self, xobj):
        """Add a companion uninitialized ensemble for comparison to verification data.

        Args:
            xobj (xarray object): Dataset/DataArray of the uninitialzed
                                  ensemble.
        """
        if isinstance(xobj, xr.DataArray):
            xobj = xobj.to_dataset()
        match_initialized_dims(self._datasets["initialized"], xobj, uninitialized=True)
        match_initialized_vars(self._datasets["initialized"], xobj)
        # Check that init is int, cftime, or datetime; convert ints or cftime to
        # datetime.
        xobj = convert_time_index(xobj, "time", "xobj[init]")
        # Check that converted/original cftime calendar is the same as the
        # initialized calendar to avoid any alignment errors.
        match_calendars(self._datasets["initialized"], xobj, kind2="uninitialized")
        datasets = self._datasets.copy()
        datasets.update({"uninitialized": xobj})
        return self._construct_direct(datasets, kind="hindcast")

    def get_observations(self):
        """Returns xarray Datasets of the observations/verification data.

        Returns:
            ``xarray`` Dataset of observations.
        """
        return self._datasets["observations"]

    def verify(
        self,
        reference=None,
        metric=None,
        comparison=None,
        dim=None,
        alignment=None,
        **metric_kwargs,
    ):
        """Verifies the initialized ensemble against observations.

        .. note::
            This will automatically verify against all shared variables
            between the initialized ensemble and observations/verification data.

        Args:
            reference (str, list of str): Type of reference forecasts to also verify against the
                observations. Choose one or more of ['uninitialized', 'persistence', 'climatology'].
                Defaults to None.
            metric (str, :py:class:`~climpred.metrics.Metric`): Metric to apply for
                verification. see `metrics </metrics.html>`_.
            comparison (str, :py:class:`~climpred.comparisons.Comparison`): How to
                compare to the observations/verification data. See
                `comparisons </comparisons.html>`_.
            dim (str, list of str): Dimension(s) to apply metric over. ``dim`` is passed
                on to xskillscore.{metric} and includes xskillscore's ``member_dim``.
                ``dim`` should contain ``member`` when ``comparison`` is probabilistic
                but should not contain ``member`` when ``comparison=e2o``. Defaults to
                ``None`` meaning that all dimensions other than ``lead`` are reduced.
            alignment (str): which inits or verification times should be aligned?

                - 'maximize': maximize the degrees of freedom by slicing ``hind`` and
                  ``verif`` to a common time frame at each lead.

                - 'same_inits': slice to a common init frame prior to computing
                  metric. This philosophy follows the thought that each lead should be
                  based on the same set of initializations.

                - 'same_verif': slice to a common/consistent verification time frame
                  prior to computing metric. This philosophy follows the thought that
                  each lead should be based on the same set of verification dates.

            **metric_kwargs (optional): arguments passed to ``metric``.

        Returns:
            Dataset with dimension skill reduced by dim containing initialized and
            reference skill(s) if specified.

        Example:
            Root mean square error (``rmse``) comparing every member with the
            verification (``m2o``) over the same verification time (``same_verifs``)
            for all leads reducing dimensions ``init`` and ``member``:

            >>> HindcastEnsemble.verify(metric='rmse', comparison='m2o',
            ...     alignment='same_verifs', dim=['init','member'])
            <xarray.Dataset>
            Dimensions:  (lead: 10)
            Coordinates:
              * lead     (lead) int32 1 2 3 4 5 6 7 8 9 10
                skill    <U11 'initialized'
            Data variables:
                SST      (lead) float64 0.08516 0.09492 0.1041 ... 0.1525 0.1697 0.1785

            Pearson's Anomaly Correlation ('acc') comparing the ensemble mean with the
            verification (``e2o``) over the same initializations (``same_inits``) for
            all leads reducing dimension ``init`` while also calculating reference
            skill for the ``persistence``, ``climatology`` and ``uninitialized``
            forecast.

            >>> HindcastEnsemble.verify(metric='acc', comparison='e2o',
            ...     alignment='same_inits', dim='init',
            ...     reference=['persistence', 'climatology' ,'uninitialized'])
            <xarray.Dataset>
            Dimensions:  (skill: 4, lead: 10)
            Coordinates:
              * lead     (lead) int32 1 2 3 4 5 6 7 8 9 10
              * skill    (skill) <U13 'initialized' 'persistence' ... 'uninitialized'
            Data variables:
                SST      (skill, lead) float64 0.9023 0.8807 0.8955 ... 0.9078 0.9128 0.9159
        """
        # Have to do checks here since this doesn't call `compute_hindcast` directly.
        # Will be refactored when `climpred` migrates to inheritance-based.
        if dim is None:
            viable_dims = list(self.get_initialized().isel(lead=0).dims) + [[]]
            raise ValueError(
                "Designate a dimension to reduce over when applying the "
                f"metric. Got {dim}. Choose one or more of {viable_dims}"
            )
        if ("member" in dim) and comparison not in ["m2o", "m2r"]:
            raise ValueError(
                "Comparison must equal 'm2o' with dim='member'. "
                f"Got comparison {comparison}."
            )
        reference = _check_valid_reference(reference)

        def _verify(
            hind,
            verif,
            hist,
            reference,
            metric,
            comparison,
            alignment,
            dim,
            **metric_kwargs,
        ):
            """Interior verify func to be passed to apply func."""
            metric, comparison, dim = _get_metric_comparison_dim(
                hind, metric, comparison, dim, kind=self.kind
            )
            forecast, verif = comparison.function(hind, verif, metric=metric)
            if metric.name == "rps":  # modify metric_kwargs for rps
                metric_kwargs = broadcast_metric_kwargs_for_rps(
                    forecast, verif, metric_kwargs
                )
            forecast = forecast.rename({"init": "time"})
            inits, verif_dates = return_inits_and_verif_dates(
                forecast,
                verif,
                alignment,
                reference=reference,
                hist=hist,
            )
            metric_over_leads = [
                _apply_metric_at_given_lead(
                    verif,
                    verif_dates,
                    lead,
                    hind=forecast,
                    hist=hist,
                    inits=inits,
                    # Ensure apply metric function returns skill and not reference
                    # results.
                    reference=None,
                    metric=metric,
                    comparison=comparison,
                    dim=dim,
                    **metric_kwargs,
                )
                for lead in forecast["lead"].data
            ]
            result = xr.concat(metric_over_leads, dim="lead", **CONCAT_KWARGS)
            result["lead"] = forecast["lead"]

            if reference is not None:
                if "member" in verif.dims:  # if broadcasted before
                    verif = verif.isel(member=0)
                for r in reference:
                    metric_over_leads = [
                        _apply_metric_at_given_lead(
                            verif,
                            verif_dates,
                            lead,
                            hind=forecast,
                            hist=hist,
                            inits=inits,
                            reference=r,
                            metric=metric,
                            comparison=comparison,
                            dim=dim,
                            **metric_kwargs,
                        )
                        for lead in forecast["lead"].data
                    ]
                    ref = xr.concat(metric_over_leads, dim="lead", **CONCAT_KWARGS)
                    ref["lead"] = forecast["lead"]
                    # fix to get no member dim for uninitialized e2o skill #477
                    if (
                        r == "uninitialized"
                        and comparison.name == "e2o"
                        and "member" in ref.dims
                    ):
                        ref = ref.mean("member")
                    result = xr.concat([result, ref], dim="skill", **CONCAT_KWARGS)
            # rename back to 'init'
            if "time" in result.dims:
                result = result.rename({"time": "init"})
            # Add dimension/coordinate for different references.
            result = result.assign_coords(skill=["initialized"] + reference)
            return result.squeeze()

        has_dataset(
            self._datasets["observations"], "observational", "verify a forecast"
        )
        if "uninitialized" in reference:
            has_dataset(
                self._datasets["uninitialized"],
                "uninitialized",
                "compute an uninitialized reference forecast",
            )
            hist = self._datasets["uninitialized"]
        else:
            hist = None

        res = self._apply_climpred_function(
            _verify,
            init=True,
            metric=metric,
            comparison=comparison,
            alignment=alignment,
            dim=dim,
            hist=hist,
            reference=reference,
            **metric_kwargs,
        )
        if self._temporally_smoothed:
            res = _reset_temporal_axis(res, self._temporally_smoothed, dim="lead")
            res["lead"].attrs = self.get_initialized().lead.attrs
        return res

    def bootstrap(
        self,
        metric=None,
        comparison=None,
        dim=None,
        alignment=None,
        reference=None,
        iterations=None,
        sig=95,
        resample_dim="member",
        pers_sig=None,
        **metric_kwargs,
    ):
        """Bootstrap with replacement according to Goddard et al. 2013.

        Args:
            metric (str, :py:class:`~climpred.metrics.Metric`): Metric to apply for
                verification, see `metrics <metrics.html>`_.
            comparison (str, :py:class:`~climpred.comparisons.Comparison`): How to
                compare to the observations/verification data, see
                `comparisons </comparisons.html>`_.
            dim (str, list of str): dimension(s) to apply metric over. ``dim`` is passed
                on to xskillscore.{metric} and includes xskillscore's ``member_dim``.
                ``dim`` should contain ``member`` when ``comparison`` is probabilistic
                but should not contain ``member`` when ``comparison='e2o'``. Defaults to
                ``None`` meaning that all dimensions other than ``lead`` are reduced.
            reference (str, list of str): Type of reference forecasts with which to
                verify. One or more of ['uninitialized', 'persistence', 'climatology'].
                If None or empty, returns no p value.
            alignment (str): which inits or verification times should be aligned?

                - 'maximize': maximize the degrees of freedom by slicing ``init`` and
                  ``verif`` to a common time frame at each lead.

                - 'same_inits': slice to a common init frame prior to computing
                  metric. This philosophy follows the thought that each lead should be
                  based on the same set of initializations.

                - 'same_verif': slice to a common/consistent verification time frame
                  prior to computing metric. This philosophy follows the thought that
                  each lead should be based on the same set of verification dates.

            iterations (int): Number of resampling iterations for bootstrapping with
                replacement. Recommended >= 500.
            sig (int, default 95): Significance level in percent for deciding whether
                uninitialized and persistence beat initialized skill.
            resample_dim (str or list): dimension to resample from. default: 'member'.

                - 'member': select a different set of members from hind
                - 'init': select a different set of initializations from hind

            pers_sig (int, default None):
                If not None, the separate significance level for persistence.
            **metric_kwargs (optional): arguments passed to ``metric``.

        Returns:
            xr.Datasets: with dimensions ``results`` (holding ``skill``, ``p``,
            ``low_ci`` and ``high_ci``) and ``skill`` (holding ``initialized``,
            ``persistence`` and/or ``uninitialized``):
                * results='verify skill', skill='initialized':
                    mean initialized skill
                * results='high_ci', skill='initialized':
                    high confidence interval boundary for initialized skill
                * results='p', skill='uninitialized':
                    p value of the hypothesis that the
                    difference of skill between the initialized and
                    uninitialized simulations is smaller or equal to zero
                    based on bootstrapping with replacement.
                * results='p', skill='persistence':
                    p value of the hypothesis that the
                    difference of skill between the initialized and persistence
                    simulations is smaller or equal to zero based on
                    bootstrapping with replacement.

        Example:
            Calculate the Pearson's Anomaly Correlation ('acc') comparing the ensemble
            mean forecast to the verification (``e2o``) over the same verification
            times (``same_verifs``) for all leads reducing dimensions ``init`` 50
            times after resampling ``member`` dimension with replacement. Also
            calculate reference skill for the ``persistence``, ``climatology``
            and ``uninitialized`` forecast and compare whether initialized skill is
            better than reference skill: Returns verify skill, probability that
            reference forecast performs better than initialized and the lower and
            upper bound of the resample.

            >>> HindcastEnsemble.bootstrap(metric='acc', comparison='e2o',
            ...     dim='init', iterations=50, resample_dim='member',
            ...     alignment='same_verifs',
            ...     reference=['persistence', 'climatology' ,'uninitialized'])
            <xarray.Dataset>
            Dimensions:  (skill: 4, results: 4, lead: 10)
            Coordinates:
              * lead     (lead) int32 1 2 3 4 5 6 7 8 9 10
              * results  (results) <U12 'verify skill' 'p' 'low_ci' 'high_ci'
              * skill    (skill) <U13 'initialized' 'persistence' ... 'uninitialized'
            Data variables:
                SST      (skill, results, lead) float64 0.9313 0.9119 ... 0.8078 0.8078
            Attributes:
                prediction_skill:            calculated by climpred https://climpred.read...
                number_of_initializations:   61
                number_of_members:           10
                alignment:                   same_verifs
                metric:                      pearson_r
                comparison:                  e2o
                dim:                         ['init']
                units:                       None
                confidence_interval_levels:  0.975-0.025
                bootstrap_iterations:        50
                p:                           probability that reference performs better t...
        """
        if iterations is None:
            raise ValueError("Designate number of bootstrapping `iterations`.")
        # TODO: replace with more computationally efficient classes implementation
        reference = _check_valid_reference(reference)
        if "uninitialized" in reference and not isinstance(
            self.get_uninitialized(), xr.Dataset
        ):
            raise ValueError("reference uninitialized requires uninitialized.")
        return bootstrap_hindcast(
            self.get_initialized(),
            self.get_uninitialized()
            if isinstance(self.get_uninitialized(), xr.Dataset)
            else None,
            self.get_observations(),
            metric=metric,
            comparison=comparison,
            dim=dim,
            alignment=alignment,
            reference=reference,
            resample_dim=resample_dim,
            sig=sig,
            iterations=iterations,
            pers_sig=pers_sig,
            **metric_kwargs,
        )

    def remove_bias(
        self,
        alignment=None,
        how="additive_mean",
        train_test_split="unfair",
        train_init=None,
        train_time=None,
        cv=False,
        **metric_kwargs,
    ):
        """Calculate and remove bias from
        :py:class:`~climpred.classes.HindcastEnsemble`.
        Bias is grouped by ``seasonality`` set via :py:class:`~climpred.options.set_options`. When wrapping xclim.sbda.adjustment use ``group`` instead.

        Args:
            alignment (str): which inits or verification times should be aligned?

                - 'maximize': maximize the degrees of freedom by slicing ``hind`` and
                  ``verif`` to a common time frame at each lead.

                - 'same_inits': slice to a common init frame prior to computing
                  metric. This philosophy follows the thought that each lead should be
                  based on the same set of initializations.

                - 'same_verif': slice to a common/consistent verification time frame
                  prior to computing metric. This philosophy follows the thought that
                  each lead should be based on the same set of verification dates.

            how (str): what kind of bias removal to perform. Defaults to 'additive_mean'. Select from:

                - 'additive_mean': correcting the mean forecast additively
                - 'multiplicative_mean': correcting the mean forecast multiplicatively
                - 'multiplicative_std': correcting the standard deviation multiplicatively
                - 'modified_quantile': `Reference <https://www.sciencedirect.com/science/article/abs/pii/S0034425716302000?via%3Dihub>`_
                - 'basic_quantile': `Reference <https://rmets.onlinelibrary.wiley.com/doi/pdf/10.1002/joc.2168>`_
                - 'gamma_mapping': `Reference <https://www.hydrol-earth-syst-sci.net/21/2649/2017/>`_
                - 'normal_mapping': `Reference <https://www.hydrol-earth-syst-sci.net/21/2649/2017/>`_
                - 'EmpiricalQuantileMapping': `Reference <https://xclim.readthedocs.io/en/stable/sdba_api.html#xclim.sdba.adjustment.EmpiricalQuantileMapping>`_
                - 'DetrendedQuantileMapping': `Reference <https://xclim.readthedocs.io/en/stable/sdba_api.html#xclim.sdba.adjustment.DetrendedQuantileMapping>`_
                - 'PrincipalComponents': `Reference <https://xclim.readthedocs.io/en/stable/sdba_api.html#xclim.sdba.adjustment.PrincipalComponents>`_
                - 'QuantileDeltaMapping': `Reference <https://xclim.readthedocs.io/en/stable/sdba_api.html#xclim.sdba.adjustment.QuantileDeltaMapping>`_
                - 'Scaling': `Reference <https://xclim.readthedocs.io/en/stable/sdba_api.html#xclim.sdba.adjustment.Scaling>`_
                - 'LOCI': `Reference <https://xclim.readthedocs.io/en/stable/sdba_api.html#xclim.sdba.adjustment.LOCI>`_

            train_test_split (str): How to separate train period to calculate the bias and test period to apply bias correction to? For a detailed description, see `Risbey et al. 2021 <http://www.nature.com/articles/s41467-021-23771-z>`_:

                - `fair`: no overlap between `train` and `test` (recommended).
                    Set either `train_init` or `train_time`.
                - `unfair`: completely overlapping `train` and `test`
                    (climpred default).
                - `unfair-cv`: overlapping `train` and `test` except for current
                    `init`, which is `left out <https://en.wikipedia.org/wiki/Cross-validation_(statistics)#Leave-one-out_cross-validation>`_
                    (set `cv='LOO'`).

            train_init (xr.DataArray, slice): Define initializations for training
                when ``alignment='same_inits/maximize'``.
            train_time (xr.DataArray, slice): Define time for training
                when ``alignment='same_verif'``.
            cv (bool or str): Only relevant when `train_test_split='unfair-cv'`. Defaults to False.

                - True/'LOO': Calculate bias by `leaving given initialization out <https://en.wikipedia.org/wiki/Cross-validation_(statistics)#Leave-one-out_cross-validation>`_
                - False: include all initializations in the calculation of bias, which
                    is much faster and but yields similar skill with a large N of
                    initializations.

            **metric_kwargs (dict): passed to ``xclim.sdba`` (including ``group``) or ``XBias_Correction``

        Returns:
            HindcastEnsemble: bias removed HindcastEnsemble.

        Example:

            Skill from raw model output without bias reduction:

            >>> HindcastEnsemble.verify(metric='rmse', comparison='e2o',
            ...     alignment='maximize', dim='init')
            <xarray.Dataset>
            Dimensions:  (lead: 10)
            Coordinates:
              * lead     (lead) int32 1 2 3 4 5 6 7 8 9 10
                skill    <U11 'initialized'
            Data variables:
                SST      (lead) float64 0.08359 0.08141 0.08362 ... 0.1361 0.1552 0.1664

            Note that this HindcastEnsemble is already bias reduced, therefore
            ``train_test_split='unfair'`` has hardly any effect. Use all
            initializations to calculate bias and verify skill:

            >>> HindcastEnsemble.remove_bias(alignment='maximize',
            ...     how='additive_mean', test_train_split='unfair'
            ... ).verify(metric='rmse', comparison='e2o', alignment='maximize',
            ... dim='init')
            <xarray.Dataset>
            Dimensions:  (lead: 10)
            Coordinates:
              * lead     (lead) int32 1 2 3 4 5 6 7 8 9 10
                skill    <U11 'initialized'
            Data variables:
                SST      (lead) float64 0.08349 0.08039 0.07522 ... 0.07305 0.08107 0.08255

            Separate initializations 1954 - 1980 to calculate bias. Note that
            this HindcastEnsemble is already bias reduced, therefore
            ``train_test_split='fair'`` worsens skill here. Generally,
            ``train_test_split='fair'`` is recommended to use for a fair
            comparison against real-time forecasts.

            >>> HindcastEnsemble.remove_bias(alignment='maximize',
            ...     how='additive_mean', train_test_split='fair',
            ...     train_init=slice('1954', '1980')).verify(metric='rmse',
            ...     comparison='e2o', alignment='maximize', dim='init')
            <xarray.Dataset>
            Dimensions:  (lead: 10)
            Coordinates:
              * lead     (lead) int32 1 2 3 4 5 6 7 8 9 10
                skill    <U11 'initialized'
            Data variables:
                SST      (lead) float64 0.132 0.1085 0.08722 ... 0.08209 0.08969 0.08732

            Wrapping methods ``how`` from `xclim <https://xclim.readthedocs.io/en/stable/sdba_api.html>`_ and providing ``group`` for ``groupby``:

            >>> HindcastEnsemble.remove_bias(alignment='same_init', group='init',
            ...     how='DetrendedQuantileMapping', train_test_split='unfair',
            ...     ).verify(metric='rmse',
            ...     comparison='e2o', alignment='maximize', dim='init')
            <xarray.Dataset>
            Dimensions:  (lead: 10)
            Coordinates:
              * lead     (lead) int32 1 2 3 4 5 6 7 8 9 10
                skill    <U11 'initialized'
            Data variables:
                SST      (lead) float64 0.09823 0.09747 0.08235 ... 0.07742 0.08115 0.08326

            Wrapping methods ``how`` from `bias_correction <https://github.com/pankajkarman/bias_correction/blob/master/bias_correction.py>`_:

            >>> HindcastEnsemble.remove_bias(alignment='same_init',
            ...     how='modified_quantile', train_test_split='unfair',
            ...     ).verify(metric='rmse',
            ...     comparison='e2o', alignment='maximize', dim='init')
            <xarray.Dataset>
            Dimensions:  (lead: 10)
            Coordinates:
              * lead     (lead) int32 1 2 3 4 5 6 7 8 9 10
                skill    <U11 'initialized'
            Data variables:
                SST      (lead) float64 0.07628 0.08293 0.08169 ... 0.1577 0.1821 0.2087
        """
        if train_test_split not in BIAS_CORRECTION_TRAIN_TEST_SPLIT_METHODS:
            raise NotImplementedError(
                f"train_test_split='{train_test_split}' not implemented. Please choose `train_test_split` from {BIAS_CORRECTION_TRAIN_TEST_SPLIT_METHODS}, see Risbey et al. 2021 http://www.nature.com/articles/s41467-021-23771-z for description and https://github.com/pangeo-data/climpred/issues/648 for implementation status."
            )

        alignment = _check_valud_alignment(alignment)

        if train_test_split in ["fair"]:
            if (
                (train_init is None)
                or not isinstance(train_init, (slice, xr.DataArray))
            ) and (alignment in ["same_inits", "maximize"]):
                raise ValueError(
                    f'When alignment="{alignment}", please provide `train_init` as xr.DataArray, e.g. `hindcast.coords["init"].slice(start, end)` or slice, e.g. `slice(start, end)`, got `train_init={train_init}`.'
                )
            if (
                (train_time is None)
                or not isinstance(train_time, (slice, xr.DataArray))
            ) and (alignment in ["same_verif"]):
                raise ValueError(
                    f'When alignment="{alignment}", please provide `train_time` as xr.DataArray, e.g. `hindcast.coords["time"].slice(start, end)` or slice, e.g. `slice(start, end)`, got `train_time={train_time}`'
                )

            if isinstance(train_init, slice):
                train_init = self.coords["init"].sel(init=train_init)
            if isinstance(train_time, slice):
                train_time = self.coords["time"].sel(time=train_time)

        if how == "mean":
            how = "additive_mean"  # backwards compatibility
        if how in ["additive_mean", "multiplicative_mean", "multiplicative_std"]:
            func = gaussian_bias_removal
        elif how in BIAS_CORRECTION_BIAS_CORRECTION_METHODS:
            func = bias_correction
        elif how in XCLIM_BIAS_CORRECTION_METHODS:
            func = xclim_sdba
        else:
            raise NotImplementedError(
                f"bias removal '{how}' is not implemented, please choose from {INTERNAL_BIAS_CORRECTION_METHODS+BIAS_CORRECTION_BIAS_CORRECTION_METHODS}."
            )

        if train_test_split in ["unfair-cv"]:
            if cv not in [True, "LOO"]:
                raise ValueError(
                    f"Please provide `cv='LOO'` when train_test_split='unfair-cv', found `cv='{cv}'`"
                )
            else:
                cv = "LOO"  # backward compatibility
            if cv not in CROSS_VALIDATE_METHODS:
                raise NotImplementedError(
                    f"cross validation method {cv} not implemented. Please choose cv from {CROSS_VALIDATE_METHODS}."
                )
            metric_kwargs["cv"] = cv

        self = func(
            self,
            alignment=alignment,
            how=how,
            train_test_split=train_test_split,
            train_init=train_init,
            train_time=train_time,
            **metric_kwargs,
        )
        return self



from .constants import GROUPBY_SEASONALITIES

OPTIONS = {
    "seasonality": "month",
    "warn_for_failed_PredictionEnsemble_xr_call": True,
    "warn_for_rename_to_climpred_dims": True,
    "warn_for_init_coords_int_to_annual": True,
    "climpred_warnings": True,
}  # defaults

_SEASONALITY_OPTIONS = frozenset(GROUPBY_SEASONALITIES)

_VALIDATORS = {
    "seasonality": _SEASONALITY_OPTIONS.__contains__,
    "warn_for_PredictionEnsemble_xr_call": lambda choice: choice
    in [True, False, "default"],
    "warn_for_rename_to_climpred_dims": lambda choice: choice
    in [True, False, "default"],
    "warn_for_init_coords_int_to_annual": lambda choice: choice
    in [True, False, "default"],
    "climpred_warnings": lambda choice: choice in [True, False, "default"],
}


class set_options:
    """Set options for climpred in a controlled context. Analogous to
    `xarray.set_options(**option) <http://xarray.pydata.org/en/stable/generated/xarray.set_options.html>`_.

    Currently supported options:

    - ``seasonality``
        - Attribute to group dimension ``groupby(f"{dim}.{seasonality}"")``.
            Used in ``reference=climatology`` and
            :py:meth:`~climpred.classes.HindcastEnsemble.remove_bias`.
        - Allowed: [``"dayofyear"``, ``"weekofyear"``, ``"month"``, ``"season"``]
        - Default: ``dayofyear``.
    - ``warn_for_failed_PredictionEnsemble_xr_call``
        - Raise UserWarning when PredictionEnsemble.xr_call,
            e.g. ``.sel(lead=[1])`` fails on one of the datasets.
        - Allowed: [True, False]
        - Default: True
    - ``warn_for_rename_to_climpred_dims``
        - Raise UserWarning when dimensions are renamed to ``CLIMPRED_DIMS`` when
            PredictionEnsemble is instantiated.
        - Allowed: [True, False]
        - Default: True
    - ``warn_for_init_coords_int_to_annual``
        - Raise UserWarning when ``init`` coordinate is of type integer and gets
            converted to annual cftime_range when PredictionEnsemble is instantiated.
        - Allowed: [True, False]
        - Default: True
    - ``climpred_warnings``
        - Overwrites all options containing ``"*warn*"``.
        - Allowed: [True, False]
        - Default: True

    Examples:
        You can use ``set_options`` either as a context manager:

        >>> kw = dict(metric='mse', comparison='e2o', dim='init',
        ...           alignment='same_verifs', reference='climatology')
        >>> with climpred.set_options(seasonality='month'):
        ...     HindcastEnsemble.verify(**kw).SST.sel(skill='climatology')
        <xarray.DataArray 'SST' (lead: 10)>
        array([0.03712573, 0.03712573, 0.03712573, 0.03712573, 0.03712573,
               0.03712573, 0.03712573, 0.03712573, 0.03712573, 0.03712573])
        Coordinates:
          * lead     (lead) int32 1 2 3 4 5 6 7 8 9 10
            skill    <U11 'climatology'

        Or to set global options:

        >>> climpred.set_options(seasonality='month')  # doctest: +ELLIPSIS
        <climpred.options.set_options object at 0x...>
    """

    def __init__(self, **kwargs):
        self.old = {}
        for k, v in kwargs.items():
            if k not in OPTIONS:
                raise ValueError(
                    "argument name %r is not in the set of valid options %r"
                    % (k, set(OPTIONS))
                )
            if k in _VALIDATORS and not _VALIDATORS[k](v):
                if k == "seasonality":
                    expected = f"Expected one of {_SEASONALITY_OPTIONS!r}"
                else:
                    expected = ""
                raise ValueError(
                    f"option {k!r} given an invalid value: {v!r}. " + expected
                )
            self.old[k] = OPTIONS[k]
        self._apply_update(kwargs)

    def _apply_update(self, options_dict):
        if (
            "climpred_warnings" in options_dict
        ):  # climpred_warnings == False overwrites all warnings options
            if not options_dict["climpred_warnings"]:
                for k in [o for o in OPTIONS.keys() if "warn" in o]:
                    options_dict[k] = False
        OPTIONS.update(options_dict)

    def __enter__(self):
        return

    def __exit__(self, type, value, traceback):
        self._apply_update(self.old)



import pytest

import climpred


@pytest.mark.xfail(
    reason="not properly implemented see https://github.com/pangeo-data/climpred/issues/605"
)
@pytest.mark.parametrize(
    "cross_validate", [False, pytest.param(True, marks=pytest.mark.xfail)]
)
def test_seasonality_remove_bias(hindcast_recon_1d_dm, cross_validate):
    """Test the climpred.set_option(seasonality) changes bias reduction. Currently fails for cross_validate bias reduction."""
    hindcast = hindcast_recon_1d_dm
    hindcast._datasets["initialized"] = (
        hindcast.get_initialized().resample(init="1MS").interpolate("linear")
    )

    alignment = "maximize"
    kw = {
        "metric": "mse",
        "comparison": "e2o",
        "dim": "init",
        "alignment": alignment,
        "reference": None,
    }

    with climpred.set_options(seasonality="dayofyear"):
        dayofyear_seasonality = hindcast.remove_bias(
            alignment=alignment, cross_validate=cross_validate
        )
    with climpred.set_options(seasonality="weekofyear"):
        weekofyear_seasonality = hindcast.remove_bias(
            alignment=alignment, cross_validate=cross_validate
        )
    assert not weekofyear_seasonality.get_initialized().identical(
        dayofyear_seasonality.get_initialized()
    )
    assert not weekofyear_seasonality.verify(**kw).identical(
        dayofyear_seasonality.verify(**kw)
    )


def test_seasonality_climatology(hindcast_recon_1d_dm):
    """Test the climpred.set_option(seasonality) changes climatology."""
    hindcast = hindcast_recon_1d_dm
    alignment = "maximize"
    kw = {
        "metric": "mse",
        "comparison": "e2o",
        "dim": "init",
        "alignment": alignment,
        "reference": "climatology",
    }
    with climpred.set_options(seasonality="dayofyear"):
        dayofyear_seasonality = hindcast.verify(**kw).sel(skill="climatology")
    with climpred.set_options(seasonality="month"):
        month_seasonality = hindcast.verify(**kw).sel(skill="climatology")
    assert not month_seasonality.identical(dayofyear_seasonality)


@pytest.mark.parametrize("option_bool", [False, True])
def test_option_warn_for_failed_PredictionEnsemble_xr_call(
    hindcast_recon_1d_dm, option_bool
):
    with climpred.set_options(warn_for_failed_PredictionEnsemble_xr_call=option_bool):
        with pytest.warns(None if not option_bool else UserWarning) as record:
            hindcast_recon_1d_dm.sel(lead=[1, 2])
        if not option_bool:
            assert len(record) == 0, print(record[0])


@pytest.mark.parametrize("option_bool", [False, True])
def test_climpred_warnings(hindcast_recon_1d_dm, option_bool):
    with climpred.set_options(warn_for_failed_PredictionEnsemble_xr_call=True):
        with climpred.set_options(climpred_warnings=option_bool):
            print(climpred.options.OPTIONS)
            with pytest.warns(UserWarning if option_bool else None) as record:
                hindcast_recon_1d_dm.sel(lead=[1, 2])
            if not option_bool:
                assert len(record) == 0, print(record[0])



import pytest

from climpred.tutorial import FILE_ALIAS_DICT, load_dataset

filepaths = list(FILE_ALIAS_DICT.keys())


@pytest.mark.parametrize("filepath", filepaths)
def test_open_dataset_locally(filepath):
    """Opens all files listed in file_alias_dict."""
    print(filepath)
    d = load_dataset(filepath)
    assert d.nbytes > 0


def test_load_datasets_empty():
    actual = load_dataset()
    assert actual is None


@pytest.mark.parametrize("cache", [False, True])
def test_load_dataset_cache(cache):
    load_dataset(cache=cache)



import numpy as np
import pytest
import xarray as xr
from scipy.stats import norm

from climpred import HindcastEnsemble
from climpred.bootstrap import bootstrap_hindcast, bootstrap_perfect_model
from climpred.comparisons import (
    NON_PROBABILISTIC_PM_COMPARISONS,
    PROBABILISTIC_HINDCAST_COMPARISONS,
    PROBABILISTIC_PM_COMPARISONS,
)
from climpred.metrics import METRIC_ALIASES, PROBABILISTIC_METRICS
from climpred.prediction import compute_hindcast, compute_perfect_model

ITERATIONS = 2

probabilistic_metrics_requiring_logical = [
    "brier_score",
    "discrimination",
    "reliability",
]

probabilistic_metrics_requiring_more_than_member_dim = [
    "rank_histogram",
    "discrimination",
    "reliability",
]

references = [
    "uninitialized",
    "persistence",
    "climatology",
    ["climatology", "uninitialized", "persistence"],
]
references_ids = [
    "uninitialized",
    "persistence",
    "climatology",
    "climatology, uninitialized, persistence",
]

xr.set_options(display_style="text")


@pytest.mark.parametrize("reference", references, ids=references_ids)
@pytest.mark.parametrize("metric", PROBABILISTIC_METRICS)
@pytest.mark.parametrize("comparison", PROBABILISTIC_HINDCAST_COMPARISONS)
def test_HindcastEnsemble_verify_bootstrap_probabilistic(
    hindcast_hist_obs_1d, metric, comparison, reference
):
    """
    Checks that HindcastEnsemble.verify() and HindcastEnsemble.bootstrap() works
    without breaking for all probabilistic metrics.
    """
    he = hindcast_hist_obs_1d.isel(lead=[0, 1], init=range(10))

    category_edges = np.array([-0.5, 0, 0.5])
    if metric in probabilistic_metrics_requiring_logical:

        def f(x):
            return x > 0

        kwargs = {"logical": f}
    elif metric == "threshold_brier_score":
        kwargs = {"threshold": 0}
    elif metric == "contingency":
        kwargs = {
            "forecast_category_edges": category_edges,
            "observation_category_edges": category_edges,
            "score": "accuracy",
        }
    elif metric == "rps":
        kwargs = {"category_edges": category_edges}
    else:
        kwargs = {}
    dim = (
        ["member", "init"]
        if metric in probabilistic_metrics_requiring_more_than_member_dim
        else "member"
    )
    # verify()
    kwargs.update(
        {
            "comparison": comparison,
            "metric": metric,
            "dim": dim,
            "reference": reference,
            "alignment": "same_verifs",
        }
    )
    actual_verify = he.verify(**kwargs)["SST"]
    not actual_verify.isnull().all()

    # bootstrap()
    actual = he.bootstrap(iterations=3, **kwargs)["SST"]
    assert "dayofyear" not in actual.coords

    if isinstance(reference, str):
        reference = [reference]
    if len(reference) == 0:
        assert not actual.sel(results="verify skill").isnull().all()
    else:
        assert (
            not actual.sel(skill="initialized", results="verify skill").isnull().all()
        )
        for skill in reference:
            actual_skill = actual.sel(skill=skill, results="verify skill")
            if metric == "crpss_es" and skill in ["climatology", "persistence"]:
                pass
            else:
                assert not actual_skill.isnull().all()


@pytest.mark.parametrize("reference", references, ids=references_ids)
@pytest.mark.parametrize("comparison", PROBABILISTIC_PM_COMPARISONS)
@pytest.mark.parametrize("metric", PROBABILISTIC_METRICS)
def test_PerfectModelEnsemble_verify_bootstrap_not_nan_probabilistic(
    perfectModelEnsemble_initialized_control, metric, comparison, reference
):
    """
    Checks that PerfectModelEnsemble.verify() and PerfectModelEnsemble.bootstrap() works without breaking for all probabilistic metrics.
    """
    pm = perfectModelEnsemble_initialized_control.isel(lead=range(3), init=range(5))
    kwargs = {
        "comparison": comparison,
        "metric": metric,
    }
    category_edges = np.array([9.5, 10.0, 10.5])
    if metric in probabilistic_metrics_requiring_logical:

        def f(x):
            return x > 10

        kwargs["logical"] = f
    elif metric == "threshold_brier_score":
        kwargs["threshold"] = 10.0
    elif metric == "contingency":
        kwargs["forecast_category_edges"] = category_edges
        kwargs["observation_category_edges"] = category_edges
        kwargs["score"] = "accuracy"
    elif metric == "rps":
        kwargs["category_edges"] = category_edges
    dim = (
        ["member", "init"]
        if metric in probabilistic_metrics_requiring_more_than_member_dim
        else "member"
    )
    kwargs["dim"] = dim

    # verify()
    actual_verify = pm.verify(**kwargs)
    assert not actual_verify.tos.isnull().all()

    # bootstrap
    kwargs["iterations"] = ITERATIONS
    kwargs["resample_dim"] = "member"
    kwargs["reference"] = reference
    actual = pm.bootstrap(**kwargs).tos
    if isinstance(reference, str):
        reference = [reference]
    if len(reference) == 0:
        assert not actual.sel(results="verify skill").isnull().all()
    else:
        for skill in reference:
            actual_skill = actual.sel(skill=skill, results="verify skill")
            if metric == "crpss_es" and skill in ["climatology", "persistence"]:
                pass
            else:
                assert not actual_skill.isnull().all()


@pytest.mark.slow
@pytest.mark.skip(reason="takes quite long")
def test_compute_hindcast_da1d_not_nan_crpss_quadratic(
    hind_da_initialized_1d, observations_da_1d
):
    """
    Checks that there are no NaNs on hindcast metrics of 1D time series.
    """
    actual = (
        compute_hindcast(
            hind_da_initialized_1d.isel(lead=[0, 1, 2], init=range(10)),
            observations_da_1d,
            comparison="m2o",
            metric="crpss",
            gaussian=False,
            dim="member",
        )
        .isnull()
        .any()
    )
    assert not actual


def test_hindcast_crpss_orientation(hind_da_initialized_1d, observations_da_1d):
    """
    Checks that CRPSS hindcast as skill score > 0.
    """
    actual = compute_hindcast(
        hind_da_initialized_1d.isel(lead=range(3)),
        observations_da_1d,
        comparison="m2o",
        metric="crpss",
        dim="member",
    )
    if "init" in actual.coords:
        actual = actual.mean("init")
    assert not (actual.isel(lead=[0, 1]) < 0).any()


# test api
# Probabilistic PM metrics dont work with non-prob PM comparison m2e and e2c
@pytest.mark.parametrize("comparison", NON_PROBABILISTIC_PM_COMPARISONS)
@pytest.mark.parametrize("metric", PROBABILISTIC_METRICS)
def test_compute_pm_probabilistic_metric_non_probabilistic_comparison_fails(
    PM_da_initialized_1d, PM_da_control_1d, metric, comparison
):
    with pytest.raises(ValueError, match=f"Probabilistic metric `{metric}` requires"):
        compute_perfect_model(
            PM_da_initialized_1d,
            PM_da_control_1d,
            comparison=comparison,
            metric=metric,
        )


@pytest.mark.parametrize("metric", ["crps"])
def test_compute_hindcast_probabilistic_metric_e2o_fails(
    hind_da_initialized_1d, observations_da_1d, metric
):
    metric = METRIC_ALIASES.get(metric, metric)
    with pytest.raises(ValueError, match=f"Probabilistic metric `{metric}` requires"):
        compute_hindcast(
            hind_da_initialized_1d,
            observations_da_1d,
            comparison="e2o",
            metric=metric,
            dim="member",
        )


def test_HindcastEnsemble_rps_terciles(hindcast_hist_obs_1d):
    actual = hindcast_hist_obs_1d.isel(lead=range(3), init=range(10)).verify(
        metric="rps",
        comparison="m2o",
        dim=["member", "init"],
        alignment="same_verifs",
        category_edges=np.array([-0.5, 0.0, 0.5, 1]),
        reference="climatology",
    )  # todo really use terciles
    assert actual.notnull().all()
    rpss = 1 - actual.sel(skill="initialized") / actual.sel(skill="climatology")
    assert rpss.isel(lead=0) > 0


def test_hindcast_verify_brier_logical(hindcast_recon_1d_ym):
    """Test that a probabilistic score requiring a binary observations and
    probability initialized inputs gives the same results whether passing logical
    as kwarg or mapping logical before for hindcast.verify()."""
    he = hindcast_recon_1d_ym.isel(lead=range(3), init=range(10))

    def logical(ds):
        return ds > 0.5

    brier_logical_passed_as_kwarg = he.verify(
        metric="brier_score",
        comparison="m2o",
        logical=logical,
        dim="member",
        alignment="same_verif",
    )
    brier_logical_mapped_before_and_member_mean = (
        he.map(logical)
        .mean("member")
        .verify(metric="brier_score", comparison="e2o", dim=[], alignment="same_verif")
    )
    brier_logical_mapped_before_no_member_mean = he.map(logical).verify(
        metric="brier_score", comparison="m2o", dim="member", alignment="same_verif"
    )
    assert (
        brier_logical_mapped_before_and_member_mean == brier_logical_passed_as_kwarg
    ).all()
    assert (
        brier_logical_mapped_before_no_member_mean == brier_logical_passed_as_kwarg
    ).all()


@pytest.mark.parametrize("aggregate", [True, False])
def test_rps_different_edges(hindcast_recon_1d_mm, aggregate):
    """Test that HindcastEnsemble.verify(metric='rps') can work with
    different category_edges for forecast and observations."""
    he = hindcast_recon_1d_mm

    if aggregate:
        he = he.smooth(dict(lead=3), how="mean")

    q = [1 / 3, 2 / 3]
    model_edges = (
        he.get_initialized()
        .groupby("init.month")
        .quantile(q=q, dim=["init", "member"])
        .rename({"quantile": "category_edge"})
    )
    obs_edges = (
        he.get_observations()
        .groupby("time.month")
        .quantile(q=q, dim="time")
        .rename({"quantile": "category_edge"})
    )
    rps = he.verify(
        metric="rps",
        dim=["member", "init"],
        alignment="maximize",
        comparison="m2o",
        category_edges=(obs_edges, model_edges),
    )
    assert "month" not in rps.dims
    assert list(rps.dims) == ["lead"]


@pytest.mark.parametrize("aggregate", [True, False])
def test_rps_one_edge(hindcast_recon_1d_mm, aggregate):
    """Test that HindcastEnsemble.verify(metric='rps') can work with same
    category_edges for forecast and observations."""
    he = hindcast_recon_1d_mm

    if aggregate:
        he = he.smooth(dict(lead=3), how="mean")

    q = [1 / 3, 2 / 3]
    obs_edges = (
        he.get_observations()
        .groupby("time.month")
        .quantile(q=q, dim="time")
        .rename({"quantile": "category_edge"})
    )
    rps = he.verify(
        metric="rps",
        dim=["member", "init"],
        alignment="maximize",
        comparison="m2o",
        category_edges=obs_edges,
    )
    assert "month" not in rps.dims
    assert list(rps.dims) == ["lead"]



import numpy as np
import pytest
import xarray as xr

import climpred
from climpred import HindcastEnsemble, PerfectModelEnsemble
from climpred.constants import HINDCAST_CALENDAR_STR, PM_CALENDAR_STR
from climpred.tutorial import load_dataset
from climpred.utils import convert_time_index

CALENDAR = PM_CALENDAR_STR.strip("Datetime").lower()

xr.set_options(display_style="text")


@pytest.fixture(autouse=True)
def add_standard_imports(
    doctest_namespace,
    hindcast_hist_obs_1d,
    hindcast_recon_3d,
    perfectModelEnsemble_initialized_control,
):
    """imports for doctest"""
    xr.set_options(display_style="text")
    doctest_namespace["np"] = np
    doctest_namespace["xr"] = xr
    doctest_namespace["climpred"] = climpred

    # always seed numpy.random to make the examples deterministic
    np.random.seed(42)

    # climpred data
    doctest_namespace["HindcastEnsemble"] = hindcast_hist_obs_1d
    doctest_namespace["HindcastEnsemble_3D"] = hindcast_recon_3d
    doctest_namespace["PerfectModelEnsemble"] = perfectModelEnsemble_initialized_control


@pytest.fixture()
def PM_ds3v_initialized_1d():
    """MPI Perfect-model-framework initialized timeseries xr.Dataset with three
    variables."""
    return load_dataset("MPI-PM-DP-1D").isel(area=1, period=-1, drop=True)


@pytest.fixture()
def PM_ds_initialized_1d(PM_ds3v_initialized_1d):
    """MPI Perfect-model-framework initialized timeseries xr.Dataset."""
    return PM_ds3v_initialized_1d.drop_vars(["sos", "AMO"])


@pytest.fixture()
def PM_da_initialized_1d(PM_ds_initialized_1d):
    """MPI Perfect-model-framework initialized timeseries xr.DataArray."""
    return PM_ds_initialized_1d["tos"]


@pytest.fixture()
def PM_da_initialized_1d_lead0(PM_da_initialized_1d):
    """MPI Perfect-model-framework initialized timeseries xr.DataArray in lead-0
    framework."""
    da = PM_da_initialized_1d
    # Convert to lead zero for testing
    da["lead"] = da["lead"] - 1
    da["init"] = da["init"] + 1
    return da


@pytest.fixture()
def PM_ds_initialized_3d_full():
    """MPI Perfect-model-framework initialized global maps xr.Dataset."""
    return load_dataset("MPI-PM-DP-3D")


@pytest.fixture()
def PM_da_initialized_3d_full(PM_ds_initialized_3d_full):
    """MPI Perfect-model-framework initialized global maps xr.Dataset."""
    return PM_ds_initialized_3d_full["tos"]


@pytest.fixture()
def PM_ds_initialized_3d(PM_ds_initialized_3d_full):
    """MPI Perfect-model-framework initialized maps xr.Dataset of subselected North
    Atlantic."""
    return PM_ds_initialized_3d_full.sel(x=slice(120, 130), y=slice(50, 60))


@pytest.fixture()
def PM_da_initialized_3d(PM_ds_initialized_3d):
    """MPI Perfect-model-framework initialized maps xr.DataArray of subselected North
    Atlantic."""
    return PM_ds_initialized_3d["tos"]


@pytest.fixture()
def PM_ds3v_control_1d():
    """To MPI Perfect-model-framework corresponding control timeseries xr.Dataset with
    three variables."""
    return load_dataset("MPI-control-1D").isel(area=1, period=-1, drop=True)


@pytest.fixture()
def PM_ds_control_1d(PM_ds3v_control_1d):
    """To MPI Perfect-model-framework corresponding control timeseries xr.Dataset."""
    return PM_ds3v_control_1d.drop_vars(["sos", "AMO"])


@pytest.fixture()
def PM_da_control_1d(PM_ds_control_1d):
    """To MPI Perfect-model-framework corresponding control timeseries xr.DataArray."""
    return PM_ds_control_1d["tos"]


@pytest.fixture()
def PM_ds_control_3d_full():
    """To MPI Perfect-model-framework corresponding control global maps xr.Dataset."""
    return load_dataset("MPI-control-3D")


@pytest.fixture()
def PM_da_control_3d_full(PM_ds_control_3d_full):
    """To MPI Perfect-model-framework corresponding control global maps xr.DataArray."""
    return PM_ds_control_3d_full["tos"]


@pytest.fixture()
def PM_ds_control_3d(PM_ds_control_3d_full):
    """To MPI Perfect-model-framework corresponding control maps xr.Dataset of
    subselected North Atlantic."""
    return PM_ds_control_3d_full.sel(x=slice(120, 130), y=slice(50, 60))


@pytest.fixture()
def PM_da_control_3d(PM_ds_control_3d):
    """To MPI Perfect-model-framework corresponding control maps xr.DataArray of
    subselected North Atlantic."""
    return PM_ds_control_3d["tos"]


@pytest.fixture()
def perfectModelEnsemble_initialized_control_3d_North_Atlantic(
    PM_ds_initialized_3d, PM_ds_control_3d
):
    """PerfectModelEnsemble with `initialized` and `control` for the North Atlantic."""
    return PerfectModelEnsemble(PM_ds_initialized_3d).add_control(PM_ds_control_3d)


@pytest.fixture()
def perfectModelEnsemble_initialized_control(PM_ds_initialized_1d, PM_ds_control_1d):
    """PerfectModelEnsemble initialized with `initialized` and `control` xr.Dataset."""
    pm = PerfectModelEnsemble(PM_ds_initialized_1d).add_control(PM_ds_control_1d)
    return pm.generate_uninitialized()


@pytest.fixture()
def perfectModelEnsemble_3v_initialized_control_1d(
    PM_ds3v_initialized_1d, PM_ds3v_control_1d
):
    """PerfectModelEnsemble 1d initialized with `initialized` and `control` xr.Dataset
    with three variables."""
    return PerfectModelEnsemble(PM_ds3v_initialized_1d).add_control(PM_ds3v_control_1d)


@pytest.fixture()
def hind_ds_initialized_1d():
    """CESM-DPLE initialized hindcast timeseries mean removed xr.Dataset."""
    ds = load_dataset("CESM-DP-SST")
    ds["SST"].attrs["units"] = "C"
    ds["init"] = ds.init.astype("int")
    return ds


@pytest.fixture()
def hind_ds_initialized_1d_cftime(hind_ds_initialized_1d):
    """CESM-DPLE initialzed hindcast timeseries with cftime initializations."""
    ds = hind_ds_initialized_1d
    ds = convert_time_index(ds, "init", "ds.init", calendar=HINDCAST_CALENDAR_STR)
    ds.lead.attrs["units"] = "years"
    return ds


@pytest.fixture()
def hind_ds_initialized_1d_lead0(hind_ds_initialized_1d):
    """CESM-DPLE initialized hindcast timeseries mean removed xr.Dataset in lead-0
    framework."""
    da = hind_ds_initialized_1d
    # Change to a lead-0 framework
    da["init"] = da["init"] + 1
    da["lead"] = da["lead"] - 1
    return da


@pytest.fixture()
def hind_da_initialized_1d(hind_ds_initialized_1d):
    """CESM-DPLE initialized hindcast timeseries mean removed xr.DataArray."""
    return hind_ds_initialized_1d["SST"]


@pytest.fixture()
def hind_ds_initialized_3d_full():
    """CESM-DPLE initialized hindcast Pacific maps mean removed xr.Dataset."""
    ds = load_dataset("CESM-DP-SST-3D")
    return ds - ds.mean("init")


@pytest.fixture()
def hind_ds_initialized_3d(hind_ds_initialized_3d_full):
    """CESM-DPLE initialized hindcast Pacific maps mean removed xr.Dataset."""
    return hind_ds_initialized_3d_full.isel(nlon=slice(0, 10), nlat=slice(0, 12))


@pytest.fixture()
def hind_da_initialized_3d(hind_ds_initialized_3d):
    """CESM-DPLE initialized hindcast Pacific maps mean removed xr.DataArray."""
    return hind_ds_initialized_3d["SST"]


@pytest.fixture()
def hist_ds_uninitialized_1d():
    """CESM-LE uninitialized historical timeseries members mean removed xr.Dataset."""
    ds = load_dataset("CESM-LE")
    ds["SST"].attrs["units"] = "C"
    # add member coordinate
    ds["member"] = range(1, 1 + ds.member.size)
    ds = ds - ds.mean("time")
    ds["SST"].attrs["units"] = "C"
    return ds


@pytest.fixture()
def hist_da_uninitialized_1d(hist_ds_uninitialized_1d):
    """CESM-LE uninitialized historical timeseries members mean removed xr.DataArray."""
    return hist_ds_uninitialized_1d["SST"]


@pytest.fixture()
def reconstruction_ds_1d():
    """CESM-FOSI historical reconstruction timeseries members mean removed
    xr.Dataset."""
    ds = load_dataset("FOSI-SST")
    ds = ds - ds.mean("time")
    ds["SST"].attrs["units"] = "C"
    return ds


@pytest.fixture()
def reconstruction_ds_1d_cftime(reconstruction_ds_1d):
    """CESM-FOSI historical reconstruction timeseries with cftime time axis."""
    ds = reconstruction_ds_1d
    ds = convert_time_index(ds, "time", "ds.init", calendar=HINDCAST_CALENDAR_STR)
    return ds


@pytest.fixture()
def reconstruction_da_1d(reconstruction_ds_1d):
    """CESM-FOSI historical reconstruction timeseries members mean removed
    xr.DataArray."""
    return reconstruction_ds_1d["SST"]


@pytest.fixture()
def reconstruction_ds_3d_full():
    """CESM-FOSI historical Pacific reconstruction maps members mean removed
    xr.Dataset."""
    ds = load_dataset("FOSI-SST-3D")
    return ds - ds.mean("time")


@pytest.fixture()
def reconstruction_ds_3d(reconstruction_ds_3d_full):
    """CESM-FOSI historical reconstruction maps members mean removed
    xr.Dataset."""
    return reconstruction_ds_3d_full.isel(nlon=slice(0, 10), nlat=slice(0, 12))


@pytest.fixture()
def reconstruction_da_3d(reconstruction_ds_3d):
    """CESM-FOSI historical reconstruction maps members mean removed
    xr.DataArray."""
    return reconstruction_ds_3d["SST"]


@pytest.fixture()
def observations_ds_1d():
    """Historical timeseries from observations matching `hind_da_initialized_1d` and
    `hind_da_uninitialized_1d` mean removed xr.Dataset."""
    ds = load_dataset("ERSST")
    ds = ds - ds.mean("time")
    ds["SST"].attrs["units"] = "C"
    return ds


@pytest.fixture()
def observations_da_1d(observations_ds_1d):
    """Historical timeseries from observations matching `hind_da_initialized_1d` and
    `hind_da_uninitialized_1d` mean removed xr.DataArray."""
    return observations_ds_1d["SST"]


@pytest.fixture()
def hindcast_recon_3d(hind_ds_initialized_3d, reconstruction_ds_3d):
    """HindcastEnsemble initialized with `initialized`, `reconstruction`(`recon`)."""
    hindcast = HindcastEnsemble(hind_ds_initialized_3d)
    hindcast = hindcast.add_observations(reconstruction_ds_3d)
    hindcast = hindcast - hindcast.sel(time=slice("1964", "2014")).mean("time").sel(
        init=slice("1964", "2014")
    ).mean("init")
    return hindcast


@pytest.fixture()
def hindcast_recon_1d_ym(hind_ds_initialized_1d, reconstruction_ds_1d):
    """HindcastEnsemble initialized with `initialized` and `recon`."""
    hindcast = HindcastEnsemble(hind_ds_initialized_1d).add_observations(
        reconstruction_ds_1d
    )
    hindcast = hindcast - hindcast.sel(time=slice("1964", "2014")).mean("time").sel(
        init=slice("1964", "2014")
    ).mean("init")
    hindcast._datasets["initialized"]["SST"].attrs = hind_ds_initialized_1d["SST"].attrs
    hindcast._datasets["observations"]["SST"].attrs = reconstruction_ds_1d["SST"].attrs
    return hindcast


@pytest.fixture()
def hindcast_hist_obs_1d(
    hind_ds_initialized_1d, hist_ds_uninitialized_1d, observations_ds_1d
):
    """HindcastEnsemble initialized with `initialized`, `uninitialzed` and `obs`."""
    hindcast = HindcastEnsemble(hind_ds_initialized_1d)
    hindcast = hindcast.add_uninitialized(hist_ds_uninitialized_1d)
    hindcast = hindcast.add_observations(observations_ds_1d)
    with xr.set_options(keep_attrs=True):
        hindcast = hindcast - hindcast.sel(time=slice("1964", "2014")).mean("time").sel(
            init=slice("1964", "2014")
        ).mean("init")
    return hindcast


@pytest.fixture()
def hindcast_recon_1d_mm(hindcast_recon_1d_ym):
    """HindcastEnsemble with initialized and reconstruction (observations) as a monthly
    observational and initialized time series (no grid)."""
    hindcast = hindcast_recon_1d_ym.sel(time=slice("1964", "1970")).sel(
        init=slice("1964", "1970")
    )
    # resample init also
    # hindcast._datasets["initialized"]['init']=xr.cftime_range(start='1964', freq='MS', periods=hindcast.coords['init'].size) # takes too long
    hindcast._datasets["initialized"].lead.attrs["units"] = "months"
    hindcast._datasets["observations"] = (
        hindcast._datasets["observations"].resample(time="1MS").interpolate("linear")
    )
    return hindcast


@pytest.fixture()
def hindcast_recon_1d_dm(hindcast_recon_1d_ym):
    """HindcastEnsemble with initialized and reconstruction (observations) as a daily
    time series (no grid)."""
    hindcast = hindcast_recon_1d_ym.sel(time=slice("1964", "1970"))
    hindcast._datasets["initialized"].lead.attrs["units"] = "days"
    hindcast._datasets["observations"] = (
        hindcast._datasets["observations"].resample(time="1D").interpolate("linear")
    )
    hindcast._datasets["observations"].attrs = hindcast_recon_1d_ym._datasets[
        "observations"
    ]
    assert "units" in hindcast.get_initialized()["SST"].attrs
    assert "units" in hindcast_recon_1d_ym.get_observations()["SST"].attrs
    assert "units" in hindcast.get_observations()["SST"].attrs
    return hindcast


@pytest.fixture()
def hindcast_S2S_Germany():
    """S2S ECMWF on-the-fly hindcasts with daily leads and weekly inits and related observations from CPC (t2m, pr) and ERA5 (gh_500)."""
    init = load_dataset("ECMWF_S2S_Germany")
    obs = load_dataset("Observations_Germany")
    return HindcastEnsemble(init).add_observations(obs)


@pytest.fixture()
def hindcast_NMME_Nino34():
    """NMME hindcasts with monthly leads and monthly inits and related IOv2 observations for SST of the Nino34 region."""
    init = load_dataset("NMME_hindcast_Nino34_sst")
    obs = load_dataset("NMME_OIv2_Nino34_sst")
    init["sst"].attrs["units"] = "C"
    obs["sst"].attrs["units"] = "C"
    return HindcastEnsemble(init).add_observations(
        obs.broadcast_like(init, exclude=("L", "S", "M"))
    )


@pytest.fixture()
def ds1():
    """Small plain multi-dimensional coords xr.Dataset."""
    return xr.Dataset(
        {"air": (("lon", "lat"), [[0, 1, 2], [3, 4, 5], [6, 7, 8]])},
        coords={"lon": [1, 3, 4], "lat": [5, 6, 7]},
    )


@pytest.fixture()
def ds2():
    """Small plain multi-dimensional coords xr.Dataset identical values but with
    different coords compared to ds1."""
    return xr.Dataset(
        {"air": (("lon", "lat"), [[0, 1, 2], [3, 4, 5], [6, 7, 8]])},
        coords={"lon": [1, 3, 6], "lat": [5, 6, 9]},
    )


@pytest.fixture()
def da1():
    """Small plain two-dimensional xr.DataArray."""
    return xr.DataArray([[0, 1], [3, 4], [6, 7]], dims=("x", "y"))


@pytest.fixture()
def da2():
    """Small plain two-dimensional xr.DataArray with different values compared to
    da1."""
    return xr.DataArray([[0, 1], [5, 6], [6, 7]], dims=("x", "y"))


@pytest.fixture()
def da_lead():
    """Small xr.DataArray with coords `init` and `lead`."""
    lead = np.arange(5)
    init = np.arange(5)
    return xr.DataArray(
        np.random.rand(len(lead), len(init)),
        dims=["init", "lead"],
        coords=[init, lead],
    )


@pytest.fixture()
def two_dim_da():
    """xr.DataArray with two dims."""
    da = xr.DataArray(
        np.vstack(
            [
                np.arange(0, 5, 1.0),
                np.arange(0, 10, 2.0),
                np.arange(0, 40, 8.0),
                np.arange(0, 20, 4.0),
            ]
        ),
        dims=["row", "col"],
    )
    return da


@pytest.fixture()
def multi_dim_ds():
    """xr.Dataset with multi-dimensional coords."""
    ds = xr.tutorial.open_dataset("air_temperature")
    ds = ds.assign(**{"airx2": ds["air"] * 2})
    return ds


@pytest.fixture()
def da_SLM():
    """Small xr.DataArray with dims `S`, `M` and  `L` for `init`, `member` and
    `lead`.
    """
    lead = np.arange(5)
    init = np.arange(5)
    member = np.arange(5)
    return xr.DataArray(
        np.random.rand(len(lead), len(init), len(member)),
        dims=["S", "L", "M"],
        coords=[init, lead, member],
    )


@pytest.fixture()
def da_dcpp():
    """Small xr.DataArray with coords `dcpp_init_year`, `member_id` and `time` as from
    `intake-esm` `hindcastA-dcpp`."""
    lead = np.arange(5)
    init = np.arange(5)
    member = np.arange(5)
    return xr.DataArray(
        np.random.rand(len(lead), len(init), len(member)),
        dims=["dcpp_init_year", "time", "member_id"],
        coords=[init, lead, member],
    )


@pytest.fixture()
def PM_ds_initialized_1d_ym_cftime(PM_ds_initialized_1d):
    """MPI Perfect-model-framework initialized timeseries xr.Dataset with init as
    cftime."""
    PM_ds_initialized_1d = convert_time_index(
        PM_ds_initialized_1d,
        "init",
        "PM_ds_initialized_1d.init",
        calendar=PM_CALENDAR_STR,
    )
    PM_ds_initialized_1d["lead"].attrs["units"] = "years"
    return PM_ds_initialized_1d


@pytest.fixture()
def PM_ds_control_1d_ym_cftime(PM_ds_control_1d):
    """To MPI Perfect-model-framework corresponding control timeseries xr.Dataset with
    time as cftime."""
    PM_ds_control_1d = convert_time_index(
        PM_ds_control_1d, "time", "PM_ds_control_1d.time", calendar=PM_CALENDAR_STR
    )
    return PM_ds_control_1d


@pytest.fixture()
def perfectModelEnsemble_initialized_control_1d_ym_cftime(
    PM_ds_initialized_1d_ym_cftime, PM_ds_control_1d_ym_cftime
):
    """PerfectModelEnsemble with MPI Perfect-model-framework initialized and control
    timeseries annual mean with cftime coords."""
    pm = PerfectModelEnsemble(PM_ds_initialized_1d_ym_cftime)
    pm = pm.add_control(PM_ds_control_1d_ym_cftime)
    return pm


@pytest.fixture()
def PM_ds_initialized_1d_mm_cftime(PM_ds_initialized_1d):
    """MPI Perfect-model-framework initialized timeseries xr.Dataset with init as
    cftime faking all inits with monthly separation in one year and lead units to
    monthly."""
    PM_ds_initialized_1d["init"] = xr.cftime_range(
        start="3004",
        periods=PM_ds_initialized_1d.init.size,
        freq="MS",
        calendar=CALENDAR,
    )
    PM_ds_initialized_1d["lead"].attrs["units"] = "months"
    return PM_ds_initialized_1d


@pytest.fixture()
def PM_ds_control_1d_mm_cftime(PM_ds_control_1d):
    """To MPI Perfect-model-framework corresponding control timeseries xr.Dataset with
    time as cftime faking the time resolution to monthly means."""
    PM_ds_control_1d["time"] = xr.cftime_range(
        start="3000", periods=PM_ds_control_1d.time.size, freq="MS", calendar=CALENDAR
    )
    return PM_ds_control_1d


@pytest.fixture()
def perfectModelEnsemble_initialized_control_1d_mm_cftime(
    PM_ds_initialized_1d_mm_cftime, PM_ds_control_1d_mm_cftime
):
    """PerfectModelEnsemble with MPI Perfect-model-framework initialized and control
    timeseries monthly mean with cftime coords."""
    pm = PerfectModelEnsemble(PM_ds_initialized_1d_mm_cftime)
    pm = pm.add_control(PM_ds_control_1d_mm_cftime)
    return pm


@pytest.fixture()
def PM_ds_initialized_1d_dm_cftime(PM_ds_initialized_1d):
    """MPI Perfect-model-framework initialized timeseries xr.Dataset with init as
    cftime faking all inits with daily separation in one year and lead units to
    daily."""
    PM_ds_initialized_1d["init"] = xr.cftime_range(
        start="3004",
        periods=PM_ds_initialized_1d.init.size,
        freq="D",
        calendar=CALENDAR,
    )
    PM_ds_initialized_1d["lead"].attrs["units"] = "days"
    return PM_ds_initialized_1d


@pytest.fixture()
def PM_ds_control_1d_dm_cftime(PM_ds_control_1d):
    """To MPI Perfect-model-framework corresponding control timeseries xr.Dataset with
    time as cftime faking the time resolution to daily means."""
    PM_ds_control_1d = PM_ds_control_1d.isel(
        time=np.random.randint(0, PM_ds_control_1d.time.size, 5000)
    )
    PM_ds_control_1d["time"] = xr.cftime_range(
        start="3000", periods=PM_ds_control_1d.time.size, freq="D", calendar=CALENDAR
    )
    return PM_ds_control_1d


@pytest.fixture()
def perfectModelEnsemble_initialized_control_1d_dm_cftime(
    PM_ds_initialized_1d_dm_cftime, PM_ds_control_1d_dm_cftime
):
    """PerfectModelEnsemble with MPI Perfect-model-framework initialized and control
    timeseries daily mean with cftime coords."""
    pm = PerfectModelEnsemble(PM_ds_initialized_1d_dm_cftime)
    pm = pm.add_control(PM_ds_control_1d_dm_cftime)
    return pm


@pytest.fixture()
def small_initialized_da():
    """Very small simulation of an initialized forecasting system."""
    inits = [1990, 1991, 1992, 1993]
    lead = [1]
    return xr.DataArray(
        np.random.rand(len(inits), len(lead)),
        dims=["init", "lead"],
        coords=[inits, lead],
    )


@pytest.fixture()
def small_verif_da():
    """Very small simulation of a verification product."""
    time = [1990, 1991, 1992, 1993, 1994]
    return xr.DataArray(np.random.rand(len(time)), dims=["time"], coords=[time])
